"""
Mask pixels at the edge of NIRCam LW
"""
import astropy.io.fits as pyfits
import glob
import numpy as np
import os

os.system('gunzip nrc_badpix_231206_*fits.gz')

nircam_edge = 8

files = glob.glob('nrc_badpix_231206_*fits')
files.sort()
for file in files:
    with pyfits.open(file, mode='update') as im:
        dq = np.zeros_like(im[0].data)
        dq[:nircam_edge,:] |= 1024
        dq[-nircam_edge:,:] |= 1024
        dq[:,:nircam_edge] |= 1024
        dq[:,-nircam_edge:] |= 1024
        im[0].data |= (dq > 0)
        
        im.flush()
        
os.system('gzip nrc_badpix_231206_*fits')



"""
Code from API for making PNG and FITS cutouts
"""

def parse_coords(coo):
    """
    Parse `?coords={ra},{dec}` input
    """
    from astropy.coordinates import SkyCoord
    import astropy.units as u
    
    cc = coo.strip()
    if ',' in cc:
        cc = cc.split(',')
    else:
        cc = cc.split(' ')
    
    if ':' in cc[0]:
        crd = SkyCoord(' '.join(cc), unit=('hour','deg'))
        ra, dec = crd.ra.degree, crd.dec.degree
    else:
        ra, dec = float(cc[0]), float(cc[1])
    
    return ra, dec

def get_hashroot():
    """
    Generate a random hash for temporary filenames
    """
    import secrets

    hash_key = secrets.token_urlsafe(16)[:8]
    return hash_key.lower().replace("-", "x")


VALID_FILTERS = [
    "g141",
    "f127m",
    "f139m",
    "f336w",
    "f140w",
    "f660n",
    "f814w",
    "g102",
    "f105w",
    "f160w",
    "f200lp",
    "f475w",
    "f775w",
    "f110w",
    "f153m",
    "f606w",
    "f438w",
    "f435w",
    "f125w",
    "f350lp",
    "f850lp",
    "f098m",
    "f658n",
    "f845m",
    "f390w",
    "f275w",
    "f763m",
    "f555w",
    "f070w-clear",
    "f090w-clear",
    "clear-f090w",
    "f115w-clear",
    "clear-f115w",
    "f150w-clear",
    "clear-f150w",
    "f200w-clear",
    "clear-f200w",
    "f277w-clear",
    "f356w-clear",
    "f410m-clear",
    "f444w-clear",
    "f150w2-clear",
    "f322w2-clear",
    "f150w2-f162m",
    "f140m-clear",
    "f182m-clear",
    "f210m-clear",
    "f187n-clear",
    "f150w2-f164n",
    "f250m-clear",
    "f300m-clear",
    "f335m-clear",
    "f360m-clear",
    "f430m-clear",
    "f460m-clear",
    "f480m-clear",
    "f560w",
    "f770w",
    "f1000w",
    "f1280w",
    "f090wn-clear",
    "f115wn-clear",
    "f150wn-clear",
    "f200wn-clear",
    "f140mn-clear",
    "f158mn-clear",
    "clearp-f277w",
    "clearp-f356w",
    "clearp-f444w",
    "clearp-f430m",
    "clearp-f480m",
    "f356w-grismr",
    "f444w-grismr",
    "f444w-f405n",
    "f444w-f466n",
    "f444w-f470n",
    "f1500w",
    "f1800w",
    "f2100w",
]


def get_thumb(output="png", remove_tiles=True, verbose=False, **kws):
    """
    Get thumbnail cutout.  This is the same function that is used behind the web-based
    cutout API, so all of the parameters are implemented here.
    """
    import os
    import glob

    import numpy as np
    from skimage.io import imsave, imread

    import matplotlib.pyplot as plt

    if False:
        plt.rcParams['backend'] = 'Agg'
        plt.ioff()

    import astropy.io.fits as pyfits
    import astropy.wcs as pywcs

    from grizli.aws import tile_mosaic, db
    from grizli.aws.tile_mosaic import cutout_from_coords, coords_to_subtile
    from grizli.pipeline import auto_script
    from grizli import utils

    CUTOUT_PREFIX = "cutout"
    
    kwargs = dict(
        ra=189.0243001,
        dec=62.1966953,
        size=10,
        scl=1,
        scale_ab=None,
        rgb_min=-0.01,
        invert=False,
        rgb_scl=[1, 1, 1],
        pl=0,
        pf=1,
        all_filters=False,
        pad=2,
        nirspec=False,
        nrs_source="magenta",
        nrs_other="pink",
        msa_other="lightblue",
        nrs_lw=1,
        nrs_alpha=0.5,
        dpi_scale=1,
    )

    for k in kwargs:
        if k in kws:
            if hasattr(kws[k], "lower"):
                if kws[k].lower() in ["false", "true"]:
                    kwargs[k] = kws[k].lower() == "true"
                elif "," in kws[k]:
                    kwargs[k] = np.array(kws[k].split(",")).astype(float).tolist()
                else:
                    try:
                        kwargs[k] = float(kws[k])
                    except ValueError:
                        kwargs[k] = kws[k]
            else:
                try:
                    kwargs[k] = float(kws[k])
                except ValueError:
                    kwargs[k] = kws[k]

    for k in ("coords", "coord"):
        if k in kws:
            coo = kws[k]
            kwargs["ra"], kwargs["dec"] = parse_coords(coo)
            break

    if "filters" in kws:
        fk = kws["filters"]
        kwargs["filters"] = []
        for f in fk.split(","):
            fsp = f.strip().lower()
            if fsp in VALID_FILTERS:
                kwargs["filters"].append(fsp)
    else:
        kwargs["filters"] = ["f814w", "f125w", "f160w"]

    kwargs["pad"] = int(kwargs["pad"])

    if "asinh" in kws:
        if kws["asinh"]:
            xkws = dict(  # filters=['f444w-clear','f277w-clear','f115w-clear'],
                rgb_scl=[2.0, 0.8, 1.0],
                norm_kwargs={
                    "stretch": "asinh",
                    "min_cut": -0.01,
                    "max_cut": 1.0,
                    "clip": True,
                    "asinh_a": 0.03,
                },
                pl=1.5,
                pf=1,
            )

            for k in xkws:
                if k not in kws:
                    kwargs[k] = xkws[k]
        else:
            kwargs["norm_kwargs"] = None
    else:
        kwargs["norm_kwargs"] = None

    if "slit" in kws:
        slit = kws["slit"]
        slit = np.array(slit.split(",")).astype(float)
        # print('Slit: ', slit)
        if len(slit) != 5:
            slit = None
    else:
        slit = None

    utils.log_comment(utils.LOGFILE, f"args: {kwargs}", verbose=verbose)
    
    hash = get_hashroot()
    base = f"{CUTOUT_PREFIX}-{hash}"
    resp = cutout_from_coords(
        output=base + "-{tile}-{filter}_{drz}",
        make_weight=(output == "fits_weight"),
        clean_subtiles=False,
        **kwargs,
    )

    utils.log_comment(utils.LOGFILE, f"cutout_from_coords: {resp}", verbose=verbose)
    
    rfix = resp[0][0].replace("f444w-f466n", "f466n-clear")
    rfix = rfix.replace("f150w2-f162m", "f162m-clear")
    rfix = rfix.replace("f150w2-f164n", "f164n-clear")
    rfix = rfix.replace("-clear", "clear").replace("clearp-", "clear")
    root = "-".join(rfix.split("-")[:-1])

    nresp = len(resp)
    fix_resp = []
    for r in resp:
        rfix = r[0].replace("f444w-f466n", "f466n-clear")
        rfix = rfix.replace("f150w2-f162m", "f162m-clear")
        rfix = rfix.replace("f150w2-f164n", "f164n-clear")
        if r[0] != rfix:
            utils.log_comment(
                utils.LOGFILE,
                f"Rename: {r[0]} to {rfix}",
                verbose=verbose
            )
            os.system(f"mv {r[0]} {rfix}")
            
            fix_resp.append((rfix, r[1]))
            if "fits" in rfix:
                with pyfits.open(rfix, mode="update") as im:
                    fi = im[0].header["FILTER"]
                    if "CLEAR" not in fi:
                        im[0].header["FILTER"] = fi.split("-")[1] + "-CLEAR"
                    im.flush()
        elif "clearp" in r[0]:
            rfix = r[0].replace("clearp-", "").replace("_dr", "-clearp_dr")
            
            utils.log_comment(
                utils.LOGFILE,
                f"Rename: {r[0]} to {rfix}",
                verbose=verbose
            )
            os.system(f"mv {r[0]} {rfix}")
            
            fix_resp.append((rfix, r[1]))
            if "fits" in rfix:
                with pyfits.open(rfix, mode="update") as im:
                    fi = im[0].header["FILTER"]
                    if fi.startswith("CLEARP"):
                        fi = fi[7:] + "-CLEARP"
                        im[0].header["FILTER"] = fi
                    im.flush()

        else:
            fix_resp.append(r)

    resp = fix_resp
    for j, f in enumerate(kwargs["filters"]):
        if f.startswith("clearp-"):
            msg = f"Rename filter {kwargs['filters'][j]} to {f.replace('clearp-','')}"
            utils.log_comment(utils.LOGFILE, msg, verbose=verbose)

            kwargs["filters"][j] = f.replace("clearp-", "") + "-clearp"
        else:
            ffix = f.lower().replace("f444w-f466n", "f466n-clear")
            ffix = ffix.replace("f150w2-f162m", "f162m-clear")
            ffix = ffix.replace("f150w2-f164n", "f164n-clear")

            if ffix != f.lower():
                msg = f"Rename filter {kwargs['filters'][j]} to {ffix}"
                utils.log_comment(utils.LOGFILE, msg, verbose=verbose)
                kwargs["filters"][j] = ffix

    if remove_tiles:
        _tile = resp[0][0].split("-")[2]
        tile_files = glob.glob(f"tile.*{_tile}.*fits")
        tile_files.sort()
        for _file in tile_files:
            utils.log_comment(utils.LOGFILE, f"rm {_file}", verbose=verbose)
            os.remove(_file)

    # COSMOS ACS tiles off by one pixel as of Apr. 2023
    for ext in ["sci", "wht"]:
        _file = f"{base}-1392-f814w_drc_{ext}.fits"
        if os.path.exists(_file):
            utils.log_comment(utils.LOGFILE, "Roll ACS by one pixel", verbose=verbose)
            
            with pyfits.open(_file, mode="update") as im:
                im[0].data = np.roll(np.roll(im[0].data, 1, axis=0), 1, axis=1)
                im.flush()

    if output.startswith("fits"):
        in_hdu = [None]
        hdul = None
        xpf = None

        for r in resp:
            for e in r:
                if e not in in_hdu:
                    im = pyfits.open(e)
                    if im[0].data.max() == 0:
                        continue

                    if hdul is None:
                        fobj = pyfits.PrimaryHDU
                    else:
                        fobj = pyfits.ImageHDU

                    in_hdu.append(e)
                    im[0].header["EXTNAME"] = im[0].header["FILTER"]
                    if "_wht" in e:
                        im[0].header["EXTVER"] = "WHT"
                    else:
                        im[0].header["EXTVER"] = "SCI"

                    # Cutout slices
                    sh = im[0].data.shape

                    imw = pywcs.WCS(im[0].header)
                    pscl = np.abs(im[0].header["CD1_1"] * 3600)

                    if xpf is None:
                        xp0 = np.squeeze(
                            imw.all_world2pix([kwargs["ra"]], [kwargs["dec"]], 0)
                        )
                        xpf = xp0 * 1.0
                        if pscl < 0.06:
                            xpf /= 2.0  # (np.array(xpf))/2.

                        # xp = np.round(xpf+1).flatten().astype(int) * 2
                        xp = np.round(xpf * 2).astype(int)

                    nn = int(np.floor(kwargs["size"] / 0.05))

                    if pscl > 0.06:
                        slx = slice((xp[0] - nn) // 2, (xp[0] + nn) // 2)
                        sly = slice((xp[1] - nn) // 2, (xp[1] + nn) // 2)
                    else:
                        slx = slice((xp[0] - nn), (xp[0] + nn))
                        sly = slice((xp[1] - nn), (xp[1] + nn))

                    slh = utils.get_wcs_slice_header(imw, slx, sly)
                    hdui = fobj(data=im[0].data[sly, slx], header=im[0].header)

                    for k in slh:
                        hdui.header[k] = slh[k]

                    if hdul is None:
                        hdul = pyfits.HDUList([hdui])
                    else:
                        hdul.append(hdui)

        if hdul is None:
            msg = "No cutout found for ra={0}, dec={1}, filters={2}"
            return msg.format(kwargs["ra"], kwargs["dec"], kwargs["filters"])
            # return find_overlapping_exposures(request)

        retfile = root + ".fits"

        for r in resp:
            for e in r:
                if e is not None:
                    utils.log_comment(utils.LOGFILE, f"rm {e}", verbose=verbose)
                    os.remove(e)

        return hdul

    # remove empty
    keep = []
    smax = 0
    pscl = 0

    for r in resp:
        for e in r:
            if e is not None:
                im = pyfits.open(e)
                if im[0].data.max() == 0:
                    os.remove(e)
                else:
                    sh = im[-0].data.shape
                    if sh[0] > smax:
                        smax = sh[0]
                        pscl = np.abs(im[0].header["CD1_1"] * 3600)

                    keep.append(e)

    if len(keep) == 0:
        msg = "No cutout found for ra={0}, dec={1}, filters={2}"
        return msg.format(kwargs["ra"], kwargs["dec"], kwargs["filters"])

    rgb_kws = dict(
        root=root,
        HOME_PATH=None,
        scl=kwargs["scl"],
        add_labels=False,
        output_dpi=72,
        get_rgb_array=True,
        show_ir=False,
        invert=kwargs["invert"],
        filters=kwargs["filters"],
        scale_ab=kwargs["scale_ab"],
        rgb_min=kwargs["rgb_min"],
        norm_kwargs=kwargs["norm_kwargs"],
        rgb_scl=kwargs["rgb_scl"],
        pl=kwargs["pl"],
        pf=kwargs["pf"],
        xsize=6,
        tick_interval=25.6,
    )

    _rgb = auto_script.field_rgb(**rgb_kws)

    if _rgb.shape[0] == smax // 2:
        rgb = np.zeros((_rgb.shape[0] * 2, _rgb.shape[1] * 2, 3), dtype=_rgb.dtype)
        for i in [0, 1]:
            for j in [0, 1]:
                rgb[i::2, j::2, :] = _rgb
    else:
        rgb = _rgb

    rsh = rgb.shape
    # Center on target position
    im = pyfits.open(keep[-1])
    sh = im[0].data.shape

    imw = pywcs.WCS(im[0].header)
    xpf = np.array(imw.all_world2pix([kwargs["ra"]], [kwargs["dec"]], 0))
    xp = np.round(xpf).flatten().astype(int)
    nn = int(np.floor(kwargs["size"] / pscl * sh[0] / smax))

    slx = slice((xp[0] - nn) * rsh[0] // sh[0], (xp[0] + nn) * rsh[0] // sh[0] + 1)
    sly = slice((xp[1] - nn) * rsh[0] // sh[0], (xp[1] + nn) * rsh[0] // sh[0] + 1)
    # print(xp, nn, pscl, smax, rgb.shape, slx, sly)

    pfile = f"{root}.rgb.png"

    # Draw slit?
    if slit is not None:

        dpi = nn * smax // sh[0]
        fig, ax = plt.subplots(1, 1, figsize=(2, 2))
        ax.axis("off")
        ax.imshow(rgb, origin="lower")
        dpi *= kwargs["dpi_scale"]

        pscale = pscl * sh[0] / smax  # utils.get_wcs_pscale(imw)
        sx = np.array([-0.5, -0.5, 0.5, 0.5, -0.5]) * slit[3] / pscale
        sy = np.array([-0.5, 0.5, 0.5, -0.5, -0.5]) * slit[4] / pscale
        theta = -slit[2] / 180 * np.pi
        _mat = np.array(
            [[np.cos(theta), -np.sin(theta)], [np.sin(theta), np.cos(theta)]]
        )

        srot = np.array([sx, sy]).T.dot(_mat)
        srot += np.squeeze(imw.all_world2pix([slit[0]], [slit[1]], 0))
        ax.plot(*srot.T, color="w", alpha=0.8, lw=1)

        ax.set_xlim(slx.start, slx.stop - 1)
        ax.set_ylim(sly.start, sly.stop - 1)
        fig.tight_layout(pad=0)
        fig.savefig(pfile, dpi=dpi)
        plt.close("all")

    elif kwargs["nirspec"] in [True]:

        if "metafile" in kws:
            metafile = kws["metafile"]
        else:
            metafile = None

        patt_num = kws["nrs_dither"] if "nrs_dither" in kws else None
        pattern_match = "" if patt_num is None else f" AND patt_num in ({patt_num})"

        cosd = np.cos(kwargs["dec"] / 180 * np.pi)

        slits = db.SQL(
            f"""select is_source, footprint, msametfl from nirspec_slits
        WHERE valid = True {pattern_match}
        AND polygon(circle(point({kwargs['ra']},{kwargs['dec']}),{kwargs['size']/3600*1.5/cosd})) @> point(ra,dec)
        """
        )

        utils.log_comment(utils.LOGFILE, f"! Slits: {len(slits)}", verbose=verbose)

        dpi = nn * smax // sh[0]
        dpi *= kwargs["dpi_scale"]

        fig, ax = plt.subplots(1, 1, figsize=(2, 2))
        ax.axis("off")
        ax.imshow(rgb, origin="lower")

        pscale = pscl * sh[0] / smax  # utils.get_wcs_pscale(imw)

        if len(slits) > 0:
            for row in slits:
                sr = utils.SRegion(row["footprint"])
                xsl, ysl = np.squeeze(imw.all_world2pix(*sr.xy[0].T, 0))
                xsl = np.append(xsl, xsl[0])
                ysl = np.append(ysl, ysl[0])
                if row["is_source"]:
                    ec = kwargs["nrs_source"]
                else:
                    ec = kwargs["nrs_other"]

                ls = "-"
                zorder = 101

                if metafile is not None:
                    if not row["msametfl"].startswith(metafile):
                        ec = kwargs["msa_other"]
                        # ls = '--'
                        zorder = 100

                ax.plot(
                    xsl,
                    ysl,
                    c=ec,
                    alpha=kwargs["nrs_alpha"],
                    lw=kwargs["nrs_lw"],
                    ls=ls,
                    zorder=zorder,
                )

        ax.set_xlim(slx.start, slx.stop - 1)
        ax.set_ylim(sly.start, sly.stop - 1)

        fig.tight_layout(pad=0)
        fig.savefig(pfile, dpi=dpi)
        plt.close("all")

    else:
        # Show all filters?
        rsub = rgb[sly, slx, :]
        if kwargs["all_filters"]:
            # Show individual filter cutouts
            ish = rsub.shape
            Nf = len(keep)
            out = np.zeros(
                (ish[0], (ish[1]) * (Nf + 1) + kwargs["pad"] * Nf, ish[2]),
                dtype=rsub.dtype,
            )
            Nx = ish[1]
            out[:, -Nx:, :] += rsub

            for ix, e in enumerate(keep):
                with pyfits.open(e) as im:
                    filters_i = [im[0].header["FILTER"].lower()]

                _ximg = auto_script.field_rgb(
                    root=root,
                    HOME_PATH=None,
                    scl=kwargs["scl"],
                    add_labels=False,
                    output_dpi=72,
                    get_rgb_array=True,
                    show_ir=False,
                    invert=True,
                    filters=filters_i,
                    scale_ab=kwargs["scale_ab"],
                    rgb_min=kwargs["rgb_min"],
                    norm_kwargs=None,
                    rgb_scl=kwargs["rgb_scl"],
                    pl=kwargs["pl"],
                    pf=kwargs["pf"],
                    xsize=6,
                    tick_interval=25.6,
                )

                ish = _ximg.shape
                if ish[0] == rsh[0] * 2:
                    # print("double", ish, rsh)
                    _img = np.zeros(rsh)
                    for i in [0, 1]:
                        for j in [0, 1]:
                            _img += _ximg[i::2, j::2, :]

                    _img = (_img / 4.0).astype(_ximg.dtype)

                elif ish[0] == rsh[0] // 2:
                    # print("half", ish, rsh)
                    _img = np.zeros(rgb.shape, dtype=_ximg.dtype)
                    for i in [0, 1]:
                        for j in [0, 1]:
                            _img[i::2, j::2, :] = _ximg
                else:
                    _img = _ximg

                utils.log_comment(
                    utils.LOGFILE,
                    f"all_filters: {filters_i}",
                    verbose=verbose
                )
                
                _img = _img[sly, slx, :]

                if _img.dtype != rgb.dtype:
                    _img = _img * 1.0 / _img.max()

                ileft = ix * (Nx + kwargs["pad"])
                out[:, ileft : ileft + Nx, :] += _img

            rsub = out

        imsave(
            pfile,
            np.clip(np.round(rsub[::-1, :, :] * 256), 0, 255).astype(np.uint8),
        )

    # retimg = send_file(pfile, mimetype='image/png')
    retimg = imread(pfile)

    # Cleanup
    utils.log_comment(utils.LOGFILE, f"rm {pfile}", verbose=verbose)
    
    os.remove(pfile)
    for f in keep:
        if os.path.exists(f):
            utils.log_comment(utils.LOGFILE, f"rm {f}", verbose=verbose)
            os.remove(f)

    return retimg



"""
Extensions of the `~pyds9.DS9` object
"""
import os

# Extend pyds9.DS9 class
try:
    # Avoid choking if pyds9 not available
    import pyds9
    ds9_obj = pyds9.DS9
except:
    print('NB: Couldn\'t `import pyds9`.  The `grizli.ds9.DS9` object will be broken')
    ds9_obj = object

def show_open_targets():
    """
    Summarize open ds9 targets
    """
    import pyds9
    import numpy as np
    
    targets = pyds9.ds9_targets()
    print('i         target frame    ra       dec       x     y ')
    for i, target in enumerate(targets):
        targ = target.split()[-1]
        ds9 = DS9(target=targ)
        ra, dec = np.asarray(ds9.get('pan fk5').split(),dtype=float)
        xi, yi = np.asarray(ds9.get('pan image').split(),dtype=float)
        
        frame = ds9.get('frame')
        print(f'{i} {targ:14} {frame:>5} {ra:8.4f} {dec:8.4f} ({xi:5.0f} {yi:5.0f})')
        
        
def get_ds9(i=-1):
    """
    Get ds9 xpa target
    
    Parameters
    ----------
    i : int
        Index of target in `~pyds9.ds9_targets`  list.
    
    Returns
    -------
    obj : `~grizli.ds9.DS9`
        DS9 object.
    """
    import pyds9
    targets = pyds9.ds9_targets()
    #print(targets)
    obj = DS9(target=targets[i].split()[-1])
    return obj
    
class DS9(ds9_obj):
    """
    Extend `~pyds9.DS9` object with convenience classes
    """

    def view(self, img, header=None, simplify_wcs=True):
        """
        From pysao
        """
        import astropy.wcs as pywcs
        from .utils import to_header
        
        if hasattr(img, 'header'):
            # FITS HDU
            self.set_np2arr(img.data)
            if not header:
                header = img.header
        else:
            self.set_np2arr(img)
        
        if header:
            if simplify_wcs:
                wcs = pywcs.WCS(header)
                simple_header = to_header(wcs)
                self.set("wcs replace", simple_header.tostring())
            else:
                self.set("wcs replace", header.tostring())
                
    def frame(self, id):
        self.set(f'frame {id}')

    def scale(self, z1, z2):
        self.set(f'scale limits {z1} {z2}')

    def set_defaults(self, match='image', verbose=False):
        """
        Match frame, set log scale
        """
        commands = f"""
        xpaset -p ds9 scale log
        xpaset -p ds9 scale limits -0.1 10
        xpaset -p ds9 cmap value 3.02222 0.647552
        xpaset -p ds9 match frame {match}
        xpaset -p ds9 frame lock {match}
        xpaset -p ds9 match colorbar
        xpaset -p ds9 lock colorbar
        xpaset -p ds9 match scale"""

        for c in commands.split('\n'):
            if 'xpaset' in c:
                try:
                    self.set(' '.join(c.split()[3:]))
                except:
                    print('xpa command failed: ', c)
                if verbose:
                    print(c)
    
    
    def match(self, match='image'):
        commands = """
        xpaset -p ds9 match frames %s
        xpaset -p ds9 frame lock %s
        """ % (match, match)

        for c in commands.split('\n'):
            if 'xpaset' in c:
                self.set(' '.join(c.split()[3:]))
    
    
    def cds_query(self, radius=1.):
        """
        Open browswer with CDS catalog query around central position
        """
        rd = self.get('pan fk5').strip()
        rdst = rd.replace('+', '%2B').replace('-', '%2D').replace(' ', '+')
        url = (f'"http://vizier.u-strasbg.fr/viz-bin/VizieR?'
               f'-c={rdst}&-c.rs={radius:.1f}"')
               
        os.system(f'open {url}')
    
    
    def eso_query(self, radius=1., dp_types=['CUBE','IMAGE'], extra=''):
        """
        Open browser with ESO archive query around central position.
        
        ``radius`` in arcmin.
        """
        ra, dec = self.get('pan fk5').strip().split()
        
        dp_type = ','.join(dp_types)
        
        url = (f'"https://archive.eso.org/scienceportal/home?'
                f'pos={ra},{dec}&r={radius/60.}&dp_type={dp_type}{extra}"')
                        
        os.system(f'open {url}')
    
    
    def mast_query(self, instruments=['WFC3','ACS','WFPC2'], max=1000):
        """
        Open browser with MAST archive query around central position
        """
        ra, dec = self.get('pan fk5').strip().split()
        if len(instruments) > 0:
            instr='&sci_instrume='+','.join(instruments)
        else:
            instr = ''
            
        url = (f'"https://archive.stsci.edu/hst/search.php?RA={ra}&DEC={dec}'
               f'&sci_aec=S{instr}&max_records={max}&outputformat=HTML_Table'
                '&action=Search"')
                
        os.system(f'open {url}')
    
    
    def alma_query(self, mirror="almascience.eso.org", extra=''):
        """
        Open browser with ALMA archive query around central position
        """
        ra, dec = self.get('pan fk5').strip().split()
    
        url = (f"https://{mirror}/asax/?result_view=observation"
               f"&raDec={ra}%20{dec}{extra}")
        os.system(f'open "{url}"')
    
    
    def show_legacysurvey(self, layer='dr8', zoom=14):
        """
        Open browser with legacysurvey.org panner around central position
        """
        ra, dec = self.get('pan fk5').strip().split()
        url = (f'"http://legacysurvey.org/viewer?ra={ra}&dec={dec}'
               f'&layer={layer}&zoom={zoom}"')
                
        os.system(f'open {url}')
        
        


import numpy as np
import astropy.units as u

KMS = u.km/u.s
FLAMBDA_CGS = u.erg/u.s/u.cm**2/u.angstrom
FNU_CGS = u.erg/u.s/u.cm**2/u.Hz

# Filter footprints
PLUS_FOOTPRINT = np.array([[0,1,0], [1,0,1], [0,1,0]]) > 0
CORNER_FOOTPRINT = (~PLUS_FOOTPRINT)
CORNER_FOOTPRINT[1,1] = False

JWST_DQ_FLAGS = [
    "DO_NOT_USE",
    "OTHER_BAD_PIXEL",
    "UNRELIABLE_SLOPE",
    "UNRELIABLE_BIAS",
    "NO_SAT_CHECK",
    "NO_GAIN_VALUE",
    "HOT",
    "WARM",
    "DEAD",
    "RC",
    "LOW_QE",
]




"""
Drizzled mosaics in tiles and subregions

Here the sky is tesselated in 4 degree patches with sizes that are 
increased slightly to be integer multiples of 512 0.1" pixels

Individual subtiles are 256 x 0.1" = 25.6"

"""
import os
from tqdm import tqdm
import numpy as np

from . import db

PIXEL_SCALE = 0.1

def define_tile_grid(a=4., phase=0.6):
    """
    Define the tile grid following the PS1 tesselation algorithm from 
    T. Budavari
    https://outerspace.stsci.edu/display/PANSTARRS/PS1+Sky+tessellation+patterns#PS1Skytessellationpatterns-ProjectioncellsintheRINGS.V3tessellation

    Paramters
    ---------
    a : float
        Tile size in degrees
        
    phase : float
       Shift the RA grid by some phase amount to better align tiles with some
       survey fields
    """
    import numpy as np
    from matplotlib import pyplot as plt
    
    from astropy import units as u
    import ligo.skymap.plot

    from astropy.coordinates import SkyCoord
    import astropy.wcs as pywcs
    
    from grizli import utils

    # bottom of first decl row at d=0
    dn = [a/2./180*np.pi]
    
    # lower decl row *centered* on d=0
    dn = [0] #a/2./180*np.pi]
    
    # shift up to align survey fields
    dn = [1./180*np.pi]
    
    theta = np.arctan(a/180*np.pi/2)*2
    dn1 = 1
    
    tn = []
    mn = []
    
    dtheta = 1.e-6
    
    while (dn1 < np.pi/2) & (dn1 > 0):
        dlo = dn[-1] - theta/2
        mni = int(np.floor(2*np.pi*np.cos(dlo)/theta)) + 1
        an = 2*np.pi / mni
        dn1 = np.arctan(np.tan(dn[-1] + theta/2.)*np.cos(an/2)) + theta/2
        ddeg = dn1 / np.pi*180
        niter = 0
        
        while ddeg < np.round(ddeg):
            niter += 1
            _theta = theta + dtheta
            dlo = dn[-1] - _theta/2
            _mni = int(np.floor(2*np.pi*np.cos(dlo)/theta)) + 1
            an = 2*np.pi / mni
            _dn1 = np.arctan(np.tan(dn[-1] + _theta/2.)*np.cos(an/2)) + _theta/2
            ddeg = _dn1 / np.pi*180
            if ddeg < np.round(ddeg):
                theta = _theta
                dn1 = _dn1
                mni = _mni
                
        mn.append(mni)
        tn.append(theta)
        
        if (dn1 < np.pi/2.) & (dn1 > 0):
            print(f'{dn[-1]/np.pi*180:.2f} {mni} {niter}')
            dn.append(dn1)
    
    if a == 2:
        dn[-1] = np.pi/2
        mn[-1] = 1
        tn[-1] = theta*2.
        
    elif (a ==4) & (dn[0] == 1./180*np.pi):
        dn[-1] = np.pi/2
        mn[-1] = 1
        tn[-1] = theta*1.5
        
    else:
        dn.append(np.pi/2)
        mn.append(1)
        tn.append(theta)

    # Double for negative 
    n = len(dn)
    strip = [s for s in range(1, n+1)]
    
    if dn[0] == 0:
        strip = [-s for s in strip[::-1]] + strip[1:]
        dn = [-dni for dni in dn[::-1]] + dn[1:]
        mn = mn[::-1] + mn[1:]
        tn = tn[::-1] + tn[1:]
    else:
        strip = [-s for s in strip[::-1]] + strip
        dn = [-dni for dni in dn[::-1]] + dn
        mn = mn[::-1] + mn
        tn = tn[::-1] + tn
    
    strip = [s for s in range(len(dn))]
    
    # Plot footprints
    import matplotlib.pyplot as plt
    from tqdm import tqdm
    
    kw = {'projection':'astro hours mollweide'}
    kw = {'projection':'astro globe'}
    kw = {'projection':'astro globe', 'center':'0d +70d'}
    
    kw = dict(projection='astro degrees zoom',
                  center='0h 86d', radius='10 deg')

    #kw = dict(projection='astro degrees zoom',
    #            center='0h 0d', radius='6 deg')
                  
    plt.close('all')
    fig, ax = plt.subplots(1,1,figsize=(8,8), 
                           subplot_kw=kw)

    #plt.title("Aitoff projection of our random data")
    
    ax.grid()
    
    nedge = 10
    step = np.linspace(0, 1, nedge)
    zeros = np.zeros(nedge)
    px = np.hstack([step, zeros+1, step[::-1], zeros])
    py = np.hstack([zeros, step, zeros+1, step[::-1]])
    
    dpix = []
    da = []
    
    names = ['tile', 'strip', 'nx', 
             'crpix1','crpix2','crval1','crval2','npix',
             'r1','d1','r2','d2','r3','d3','r4','d4']
    
    rows = []
    tileid = 0
    
    for j in tqdm(range(len(dn))):
                    
        ai = 2*np.tan(tn[j]/2)*180/np.pi
        npix = ai*3600/PIXEL_SCALE
        npixr = int(npix // 512 + 1)*512
        ai = npixr*PIXEL_SCALE/3600
        da.append(ai)
        dpix.append(npixr)
        
        ddeg = dn[j]/np.pi*180
        
        h, w = utils.make_wcsheader(ra=0, dec=ddeg, size=ai*3600, 
                                    pixscale=PIXEL_SCALE)
        h['CRPIX1'] += 0.5
        h['CRPIX2'] += 0.5
        
        ras = np.linspace(0, 2*np.pi, np.maximum(mn[j]+1, 1))[:-1]
        
        # Shift to avoid straddling ra=0
        if mn[j] > 1:
            ras += ras[1]*phase
        
        col = 'k'
        
        for ir, r in enumerate(ras):
            h['CRVAL1'] = r/np.pi*180
            h['CRVAL2'] = ddeg
            w = pywcs.WCS(h)
            fp = w.calc_footprint()
            
            tileid += 1
            row = [tileid, strip[j], ir+1]
            for k in ['CRPIX1','CRPIX2','CRVAL1','CRVAL2']:
                row.append(h[k])
            
            row.append(npixr)
            row.extend(fp.flatten().astype(np.float32).tolist())
            rows.append([d for d in row])
            
            #fpx, fpy = w.calc_footprint().T
            #fpx = np.append(fpx, fpx[0])
            #fpy = np.append(fpy, fpy[0])
            fpx, fpy = w.all_pix2world(px*npixr, py*npixr, 0)
            
            if ir > 0:
                pl = ax.plot_coord(SkyCoord(fpx, fpy, unit=('deg','deg')), 
                          color=col, alpha=0.5, linewidth=1)#, marker='.')
            else:
                pl = ax.plot_coord(SkyCoord(fpx, fpy, unit=('deg','deg')), 
                                   alpha=0.5, linewidth=1)#, marker='.')
                col = pl[0].get_color()
            
            if (j == 42) & (ir == 1):
                for ik in range(5):
                    for jk in range(5):
                        slx = slice(ik*256, (ik+1)*256)
                        sly = slice(jk*256, (jk+1)*256)
                        
                        wsl = w.slice((sly, slx))
                        fpx, fpy = wsl.all_pix2world(px*256, py*256, 0)
                        pl = ax.plot_coord(SkyCoord(fpx, fpy, 
                                                    unit=('deg','deg')), 
                                  color='k', alpha=0.5, linewidth=1)
    
    tab = utils.GTable(names=names, rows=rows)
    
    return tab
    
    ###### Checking
    df = tab.to_pandas()
    df.to_sql('mosaic_tiles', db._ENGINE, index=False, if_exists='fail', 
              method='multi')
    
    # Look at HST fields
    
    avg_coo = db.SQL("""SELECT parent, count(parent), 
                               avg(ra) as ra, avg(dec) as dec
                        FROM assoc_table where ra > 0
                        GROUP by parent order by count(parent)""")
    
    exp = db.SQL('SELECT assoc, crval1, crval2, footprint FROM exposure_files')
    coo = SkyCoord(exp['crval1'], exp['crval2'], unit=('deg','deg'))
    
    # old
    avg_coo = db.SQL("""SELECT parent, count(parent), 
                               avg(ra) AS ra, avg(dec) as dec 
                        FROM exposure_log
                        WHERE ra > 0 
                          AND awspath not like '%%grizli-cosmos-v2%%'
                        GROUP by parent order by count(parent)""")
                        
    exp = db.SQL("""SELECT parent, ra, dec 
                    FROM exposure_log 
                    WHERE ra > 0 
                      AND awspath NOT LIKE '%%grizli-cosmos-v2%%'""")
                      
    coo = SkyCoord(exp['ra'], exp['dec'], unit=('deg','deg'))
    
    ra, dec = 150.1, 2.2 # cosmos
    ra, dec = 150.1, 2.2 # cosmos
    
    test = avg_coo['count'] > 20
    test = avg_coo['count'] > 150
    
    for k in np.where(test)[0]:
        print(avg_coo['parent'][k])
        ra, dec = avg_coo['ra'][k], avg_coo['dec'][k]
        
        ctab = utils.GTable()
        ctab['ra'] = [ra]
        ctab['dec'] = dec
        idx, dr = ctab.match_to_catalog_sky(tab, other_radec=('crval1','crval2'))
    
        kw = dict(projection='astro degrees zoom',
                      center=f'{ra}d {dec}d', radius='9 deg')
                  
        #plt.close('all')
        fig, ax = plt.subplots(1,1,figsize=(8,8), 
                               subplot_kw=kw)
    
        corners = np.array([np.array(tab[c]) for c in ['r1','d1','r2','d2','r3','d3','r4','d4']])[:,dr < 10*u.deg].T
    
        ax.plot_coord(coo, linestyle='None', marker='.')
    
        for c in corners:
            cc = np.append(c, c[:2])
            ci = SkyCoord(cc[0::2], cc[1::2], unit=('deg','deg'))
            ax.plot_coord(ci, color='k', alpha=0.5, linewidth=1)
        
        ax.grid()
        ax.set_title(avg_coo['parent'][k])


def add_exposure_batch():
    """
    Add a bunch of exposures to the `mosaic_tiles_exposures` table
    """    
    import astropy.table
    import astropy.table
    from tqdm import tqdm
    
    from grizli.aws.tile_mosaic import add_exposure_to_tile_db
    
    filters = db.SQL("""SELECT filter, count(filter) 
                          FROM exposure_files 
                        GROUP BY filter 
                        ORDER BY count(filter)""")
    
    ii = len(filters)-1
    filt = 'F814W'
    
    for ii, filt in enumerate(filters['filter']):
        print(f"{ii} / {len(filters)} {filt} {filters['count'][ii]}")
        
        exp = db.SQL(f"""SELECT eid, assoc, dataset, extension, filter, 
                                sciext, crval1 as ra, crval2 as dec, footprint
                           FROM exposure_files
                           WHERE filter = '{filt}'""")
    
        tiles = db.SQL('select * from mosaic_tiles')
    
        res = [add_exposure_to_tile_db(row=exp[i:i+1], tiles=tiles)
                  for i in tqdm(range(len(exp)))]
        
        for j in range(len(res))[::-1]:
            if res[j] is None:
                print(f"Pop {exp['assoc'][j]} {j}")
                res.pop(j)
                
        db.execute(f"""DELETE from mosaic_tiles_exposures t
                USING exposure_files e
                WHERE t.expid = e.eid
                AND filter = '{filt}'
                """)
        
        N = 100
        for j in tqdm(range(len(res)//N+1)):
            sl = slice(j*N, (j+1)*N)
            #print(j, sl.start, sl.stop)
            
            tab = astropy.table.vstack(res[sl])
             
            df = tab.to_pandas()
            df.to_sql('mosaic_tiles_exposures', db._ENGINE, index=False, 
                      if_exists='append', method='multi')
        #
        # Table updates
        if 0:
            db.execute('ALTER TABLE exposure_files ADD COLUMN eid SERIAL PRIMARY KEY;')

            db.execute('GRANT ALL PRIVILEGES ON ALL TABLEs IN SCHEMA public TO db_iam_user')
            db.execute('GRANT ALL PRIVILEGES ON ALL SEQUENCES IN SCHEMA public TO db_iam_user')
            db.execute('GRANT SELECT ON ALL TABLEs IN SCHEMA public TO readonly')
            db.execute('GRANT SELECT ON ALL SEQUENCES IN SCHEMA public TO readonly')
            
            db.execute('ALTER TABLE assoc_table ADD COLUMN aid SERIAL PRIMARY KEY;')
            db.execute('CREATE INDEX on exposure_files (eid)')
            db.execute('CREATE INDEX on exposure_files (eid,filter)')
            db.execute('CREATE INDEX on mosaic_tiles_exposures (expid)')

            db.execute('CREATE INDEX on mosaic_tiles_exposures (tile, subx, suby)')
        
def make_exposure_maps():
    """
    """

    from grizli import utils
    from grizli.aws import tile_mosaic
    from grizli.aws import db
    
    filt = 'F160W'
    
    ra, dec, rsize, name =  53.14628, -27.814, 20, 'gds'
    #ra, dec, rsize, name = 189.22592,  62.24586, 20, 'gdn'
    # ra, dec, rsize, name = 214.95, 52.9, 20, 'egs'
    ra, dec, rsize, name = 150.11322, 2.24068, 48, 'cos'
    ra, dec, rsize, name = 34.34984, -5.18390, 30, 'uds'
    
    ra, dec, rsize, name = 177.40124999999998, 22.39947222222, 12, 'macs1149'
    ra, dec, rsize, name = 157.30641, 26.39197, 12, 'sdss1029'
    ra, dec, rsize, name = 215.93, 24.07, 15, 'macs1423'
    ra, dec, rsize, name = 64.39, -11.91, 15, 'macs0417'
    ra, dec, rsize, name = 3.5301941, -30.3854942, 15, 'abell2744'
    extra = ''
    
    fig, tab = tile_mosaic.exposure_map(ra, dec, rsize, name, 
                                        filt=filt.upper(), s0=18, 
                                        extra=extra)
    fig.tight_layout(pad=0.5)
    fig.savefig('/tmp/map.png')
    fig.tight_layout(pad=0.5)
    fig.savefig('/tmp/map.png')
    
    from grizli.aws import db
    mf = db.SQL("""
    SELECT * from mosfire_extractions natural join mosfire_datemask
    """)
    
    # MF with HST
    # SQL = f"""SELECT m.file, count(m.file)
    #         FROM mosfire_extractions m, exposure_files e
    #         WHERE ('(' || ra_targ || 
    #                 ', ' || dec_targ || ')')::point
    #                 <@ polygon(e.footprint)
    #         GROUP BY m.file"""
    # 
    # res = db.SQL(SQL) 
    
    ###############
    
def find_mosaic_segments(bs=16):
    """
    
    Find "segments" of connected subimages within a tile
    
    bs : bin size relative to 256*0.1" subimages
    
    """
    from scipy.ndimage import label

    from grizli import utils
    from grizli.aws import db
    
    cells = db.SQL(f"""SELECT tile, subx, suby, subra, subdec, filter, 
                              assoc, dataset, exptime
            FROM mosaic_tiles_exposures t, exposure_files e
            WHERE t.expid = e.eid
            """)
    
    cells['segment'] = 0
    
    un = utils.Unique(cells['tile'])
    ns = 0
    
    for t in un.values:
        uni = np.where(un[t])[0]
        
        # img = np.zeros((cells['suby'][uni].max()+1, 
        #                 cells['subx'][uni].max()+1), dtype=bool)
        # img[cells['suby'][uni], cells['subx'][uni]] = True
    
        bs = 16
    
        img = np.zeros((cells['suby'][uni].max()//bs+1, 
                        cells['subx'][uni].max()//bs+1), dtype=bool)
        img[cells['suby'][uni]//bs, cells['subx'][uni]//bs] = True
    
        labeled_array, num_features = label(img)
    
        cells['segment'][uni] = labeled_array[cells['suby'][uni]//bs, 
                                              cells['subx'][uni]//bs] + ns
        
        print(f'tile: {t}  npatch: {num_features}')
        
        ns += num_features

    
    names = ['tile','patch','ra','dec','jname','count','filter',
             'xmin','xmax','ymin','ymax',
             'rmin','rmax','dmin','dmax','status','mtime']
    
    un = utils.Unique(cells['segment'])
    rows = []
    
    import astropy.time
    from tqdm import tqdm
    
    for s in tqdm(un.values):
        uni = np.where(un[s])[0]
        
        unf = utils.Unique(cells['filter'][uni], verbose=False)
        ra = np.mean(cells['subra'][uni])
        dec = np.mean(cells['subdec'][uni])
        jname = utils.radec_to_targname(ra=ra, dec=dec)
        for f in unf.values:
            unfi = cells[uni][unf[f]]
            
            row = [unfi['tile'][0], unfi['segment'][0], 
                   ra, dec, jname, len(unfi), f, 
                   unfi['subx'].min(), unfi['subx'].max(), 
                   unfi['suby'].min(), unfi['suby'].max(), 
                   unfi['subra'].min(), unfi['subra'].max(), 
                   unfi['subdec'].min(), unfi['subdec'].max(),
                   0, astropy.time.Time.now().mjd]
            rows.append(row)
    
    patches = utils.GTable(names=names, rows=rows)
    
    i = 0
    t0 = 0
    
    for i in range(30):
        t = patches['tile'][i]

        os.system(f"""aws s3 sync s3://grizli-mosaic-tiles/Tiles/{t}/ ./  --exclude "*" --include "*{patches['filter'][i].lower()}*sci.fits"  """)
    
        build_mosaic_from_subregions(root=patches['jname'][i], tile=t, 
                                     files=None, 
                                     filter=patches['filter'][i].lower())
        #
        os.system(f'rm tile.{t0:04d}*')
        


def get_axis_center_coord(ax):
    """
    Get sky coords at the center of symap axis
    """
    
    tr = ax.get_transform('world').inverted()
    return tr.transform((np.mean(ax.get_xlim()), np.mean(ax.get_ylim())))


def exposures_in_axis(ax, extra_where=""):
    """
    Query exposure_files
    """
    coo = get_axis_center_coord(ax)
    
    point = f"point '({coo[0]}, {coo[1]})'"
    
    res = db.SQL(f"""SELECT file, filter, assoc from exposure_files 
    WHERE polygon(footprint) @> {point}
    {extra_where}
    ORDER BY assoc
    """)
    return res
    
    
def exposure_map(ra, dec, rsize, name, filt='F160W', s0=16, cmap='viridis', figsize=(6,6), show_tiles=True, res=None, alpha=1., ec='None', vmin=None, vmax=None, extra=''):
    """
    Make an exposure map from a database query
    """
    import ligo.skymap.plot
    from matplotlib import pyplot as plt
    import numpy as np
    
    from astropy.coordinates import SkyCoord
    from grizli.aws import tile_mosaic
    from grizli import utils
    
    cosd = np.cos(dec/180*np.pi)
    
    if res is None:
        res = db.SQL(f"""SELECT tile, subx, suby, subra, subdec, filter, 
                            COUNT(filter) as nexp, 
                            SUM(exptime) as exptime,
                            MIN(expstart) as tmin, 
                            MAX(expstart) as tmax 
            FROM mosaic_tiles_exposures t, exposure_files e
            WHERE t.expid = e.eid
            AND filter = '{filt}'
            {extra}
            AND ABS(subra - {ra})*{cosd} < {rsize/60}
            AND ABS(subdec - {dec}) < {rsize/60}
            GROUP BY tile, subx, suby, subra, subdec, filter
            """)

    kw = dict(projection='astro hours zoom',
              center=f'{ra}d {dec}d', radius=f'{rsize} arcmin')
    
    s = np.maximum(s0*18/rsize, 1)
    
    fig, ax = plt.subplots(1,1,figsize=figsize, 
                           subplot_kw=kw)
    
    ax.grid()
    
    coo = SkyCoord(res['subra'], res['subdec'], unit=('deg','deg'))
    ax.scatter(coo.ra, coo.dec, 
               c=np.log10(res['exptime']),
               marker='s', s=s, 
               cmap=cmap, alpha=alpha, ec=ec, 
               vmin=vmin, vmax=vmax, 
               transform=ax.get_transform('world'))
               
    #ax.scatter_coord(coo, c=np.log10(res['exptime']), marker='s', s=s, 
    #                 cmap=cmap, alpha=alpha, ec=ec, vmin=vmin, vmax=vmax)
        
    if show_tiles:
        for t in np.unique(res['tile']):
            twcs = tile_mosaic.tile_wcs(t)
            coo = SkyCoord(*twcs.calc_footprint().T, unit=('deg','deg'))
            ax.plot_coord(coo, color='r', linewidth=1.2, alpha=0.5)

        # Tile labels
        un = utils.Unique(res['tile'], verbose=False)

        dp = 512*0.1/3600
        rp = dp/cosd
    
        for t in un.values:
            c = (res['subra'][un[t]].min(), res['subdec'][un[t]].min(), 
                 res['subra'][un[t]].max(), res['subdec'][un[t]].max())
            #
            cp = (res['subx'][un[t]].min(), res['suby'][un[t]].min(), 
                 res['subx'][un[t]].max(), res['suby'][un[t]].max())
        
            rc = np.array([c[0]-rp, c[0]-rp, c[2]+rp, c[2]+rp, c[0]-rp])
            dc = np.array([c[1]-dp, c[3]+dp, c[3]+dp, c[1]-dp, c[1]-dp])
        
            label = f'{t:04d}: {cp[0]:03d} - {cp[2]:03d}, {cp[1]:03d} - {cp[3]:03d}'
        
            ax.plot_coord(SkyCoord(rc, dc, unit=('deg','deg')),
                          alpha=0.8, label=label, linewidth=1.2)
    
        ax.legend()
    
    ax.set_xlabel('R.A.')
    ax.set_ylabel('Dec.')
    ax.set_title(f'{name} - {filt}')
    
    #fig.tight_layout(pad=0.8)
    
    return fig, res


TILES = None
TILE_WCS = {}

def coords_to_subtile(ra=189.0243001, dec=62.19669, size=0):
    """
    Get tile/subtile associated with sky coordinates
    """
    from grizli import utils
    from astropy.coordinates import SkyCoord
    import astropy.units as u
    
    global TILES
    
    if TILES is None:
        print('Initialize TILES')
        
        if os.path.exists('TILES.csv'):
            TILES = utils.read_catalog('TILES.csv')
        else:
            TILES = db.SQL('select * from mosaic_tiles')
        
        TILES['nsub'] = TILES['npix'] // 256
        TILES['coo'] = SkyCoord(TILES['crval1'], TILES['crval2'], 
                                unit=('deg','deg'))
    
    dr = TILES['coo'].separation(SkyCoord(ra, dec, unit=('deg','deg')))
    in_tile = dr < (2*np.sqrt(2)*u.deg)
    xTILES = TILES[in_tile]
    
    tp = np.array([np.squeeze(tile_wcs(t).all_world2pix([ra], [dec], 
                                                        0)).flatten() 
                   for t in xTILES['tile']]).T
    
    # tp = np.round(tp)
    
    in_tile = ((tp > 0) & (tp < xTILES['npix'])).sum(axis=0) == 2
    subt = xTILES[in_tile]
    
    subt['fsubx'], subt['fsuby'] = tp[:,in_tile] / 256
    subt['subx'] = subt['fsubx'].astype(int)
    subt['suby'] = subt['fsuby'].astype(int)
    
    ds = size/PIXEL_SCALE/256
    subt['xmin'] = np.clip((subt['fsubx'] - ds), 0, subt['npix']).astype(int)
    subt['xmax'] = np.clip((subt['fsubx'] + ds), 0, subt['npix']).astype(int)
    subt['ymin'] = np.clip((subt['fsuby'] - ds), 0, subt['npix']).astype(int)
    subt['ymax'] = np.clip((subt['fsuby'] + ds), 0, subt['npix']).astype(int)
    
    subt['ncut'] = (subt['xmax'] - subt['xmin'] + 1)
    subt['ncut'] *= (subt['ymax'] - subt['ymin'] + 1)
    so = np.argsort(subt['ncut'])
    
    return subt[so]


def cutout_from_coords(output='mos-{tile}-{filter}_{drz}', ra=189.0243001, dec=62.1966953, size=10, filters=['F160W'], theta=0, clean_subtiles=False, send_to_s3=False, make_weight=True, **kwargs): 
    
    subt = coords_to_subtile(ra=ra, dec=dec, size=size)[0]
    
    ll = (subt['xmin'], subt['ymin'])
    ur = (subt['xmax'], subt['ymax'])
    
    resp = []
    for filt in filters:
        ri = build_mosaic_from_subregions(root=output, 
                                     tile=subt['tile'], files=None, 
                                     filter=filt, 
                                     ll=ll, ur=ur,
                                     clean_subtiles=clean_subtiles, 
                                     make_weight=make_weight,
                                     send_to_s3=send_to_s3)
        resp.append(ri)
        
    return resp


def add_exposure_to_tile_db(dataset='ibev8xubq', sciext=1, tiles=None, row=None):
    """
    Find subtiles that overlap with an exposure in the `exposure_files`
    table
    """
    import astropy.table
    import astropy.units as u
    import astropy.wcs as pywcs
    
    import numpy as np
    from grizli import utils
    
    global TILES
    
    if TILES is None:
        TILES = db.SQL('select * from mosaic_tiles')
        
    if tiles is None:
        tiles = TILES

    if row is None:
        row = db.SQL(f"""SELECT eid, assoc, dataset, extension, filter, 
                                sciext, crval1 as ra, crval2 as dec, footprint 
                           from exposure_files
                           where filter = 'F160W' 
                             and dataset = '{dataset}' 
                             AND sciext={sciext}""")
    
    if len(row) == 0:
        print(f'No exposure data found for dataset={dataset} sciext={sciext}')
        return None
                    
    idx, dr = row.match_to_catalog_sky(tiles, other_radec=('crval1','crval2'))
    
    tix = np.where(dr < 4*np.sqrt(2)*u.deg)[0]
    #tix = np.where(dr < 4*u.deg)[0]
    
    exp_poly = None
    for fp in row['footprint']:
        sr = utils.SRegion(fp, wrap=False)
        sra = sr.xy[0][:,0]
        if (sra.min() < 10) & (sra.max() > 350):
            sra[sra > 350] -= 360

        sr.xy[0][:,0] = sra

        for p, s in zip(sr.get_patch(alpha=0.5, color='k'), sr.shapely):
            if exp_poly is None:
                exp_poly = s
            else:
                exp_poly = exp_poly.union(s)
                
            # ax.add_patch(p)
    
    sbuff = utils.SRegion(np.array(exp_poly.buffer(1./60).boundary.xy).T)
    
    tabs = []
    
    for t in tix:
        
        h, w = utils.make_wcsheader(ra=tiles['crval1'][t], 
                                    dec=tiles['crval2'][t],
                                    size=tiles['npix'][t]*PIXEL_SCALE, 
                                    pixscale=PIXEL_SCALE)
        
        h['CRPIX1'] += 0.5
        h['CRPIX2'] += 0.5
        h['LATPOLE'] = 0.
        
        w = pywcs.WCS(h)
        wfp = w.calc_footprint()

        wra = wfp[:,0]
        if (wra.min() < 10) & (wra.max() > 350):
            if sr.centroid[0][0] < 10:
                wra[wra > 350] -= 360
            else:
                wra[wra < 10] += 360

        wfp[:,0] = wra
        srt = utils.SRegion(wfp, wrap=False)

        if not sr.shapely[0].intersects(srt.shapely[0]):
            continue
        
        nsub = tiles['npix'][t]//256
        step = np.arange(nsub)
        px, py = np.meshgrid(step, step)
        px = px.flatten()
        py = py.flatten()
        rd = w.all_pix2world(px*256+128, py*256+128, 0)
        pts = np.array([rd[0], rd[1]]).T
        test = sbuff.path[0].contains_points(pts)
        tw = np.where(test)[0]
        if test.sum() == 0:
            continue
        
        for j, xi, yi in zip(tw, px[tw], py[tw]):
            wsl = w.slice((slice(yi*256, (yi+1)*256), 
                           slice(xi*256, (xi+1)*256)))
            sw = utils.SRegion(wsl.calc_footprint())
            test[j] = sw.shapely[0].intersects(exp_poly)
        
        if test.sum() == 0:
            continue
            
        tmatch = utils.GTable()
        tmatch['tile'] = [tiles['tile'][t]] * test.sum()
        tmatch['subx'] = px[test]
        tmatch['suby'] = py[test]
        tmatch['subra'] = rd[0][test]
        tmatch['subdec'] = rd[1][test]
        tmatch['expid'] = np.unique(row['eid'])[0]
        # tmatch['eassoc'] = np.unique(row['assoc'])[0]
        # tmatch['edataset'] = np.unique(row['dataset'])[0]
        # tmatch['eext'] = np.unique(row['sciext'])[0]
        tmatch['in_mosaic'] = 0
        tabs.append(tmatch)
        
    if len(tabs) > 0:
        tmatch = astropy.table.vstack(tabs)
        return tmatch
    else:
        return None


def tile_wcs(tile):
    """
    Compute tile WCS
    """
    import astropy.wcs as pywcs
    from grizli import utils
    from grizli.aws import db
    
    global TILES
    if TILES is None:
        TILES = db.SQL('select * from mosaic_tiles')
    
    if tile not in TILES['tile']:
        print(f'{tile} not in `mosaic_tiles`')
        return None
    
    if tile in TILE_WCS:
        return TILE_WCS[tile]
            
    row = TILES[TILES['tile'] == tile]
        
    # row = db.SQL(f"""SELECT crval1, crval2, npix
    #                    FROM mosaic_tiles
    #                   WHERE tile={tile}""")

    t = 0
    h, w = utils.make_wcsheader(ra=row['crval1'][t], dec=row['crval2'][t],
                                size=row['npix'][t]*PIXEL_SCALE,
                                pixscale=PIXEL_SCALE)
    
    # h['CRPIX1'] += 0.5
    # h['CRPIX2'] += 0.5
    h['LATPOLE'] = 0.
    
    wcs = pywcs.WCS(h)
    wcs.pscale = PIXEL_SCALE
    
    TILE_WCS[tile] = wcs
    
    return wcs
    
    
def tile_subregion_wcs(tile, subx, suby):
    """
    Compute WCS for a tile subregion
    """
    
    twcs = tile_wcs(tile)
    
    sub_wcs = twcs.slice((slice(suby*256, (suby+1)*256), 
                          slice(subx*256, (subx+1)*256)))
    sub_wcs.pscale = PIXEL_SCALE
    return sub_wcs


#
def get_lambda_client(region_name='us-east-1'):
    """
    Get boto3 client in same region as HST public dataset
    """
    import boto3
    session = boto3.Session()
    client = session.client('lambda', region_name=region_name)
    return client


def send_event_lambda(event, verbose=True, client=None, func='grizli-mosaic_tile'):
    """
    Send a single event to AWS lambda
    """
    import time
    import os
    import yaml

    import numpy as np
    import boto3
    import json

    if client is None:
        client = get_lambda_client(region_name='us-east-1')

    if verbose:
        print('Send event to {0}: {1}'.format(func, event))

    response = client.invoke(FunctionName=func,
                             InvocationType='Event', LogType='Tail',
                             Payload=json.dumps(event))


def count_locked():
    """
    Count number of distinct locked tiles that could still be runing
    on aws lambda
    """
    tiles = db.SQL("""SELECT tile, subx, suby, filter, count(filter)
                      FROM mosaic_tiles_exposures t, exposure_files e
                               WHERE t.expid = e.eid AND tile != 1183
                               AND in_mosaic=9
                               GROUP BY tile, subx, suby, filter
                               ORDER BY count(filter) DESC""")
    
    return len(tiles), tiles


def reset_locked():
    """
    Reset in_mosaic 9 > 0 for tiles that may have timed out
    """
    cmd = """UPDATE mosaic_tiles_exposures
             SET in_mosaic = 0 WHERE in_mosaic = 9"""
             
    db.execute(cmd)


def get_subtile_status(tile=2530, subx=522, suby=461, **kwargs):
    """
    `in_mosaic` status of all entries of a subtile
    """
    resp = db.SQL(f"""SELECT tile, subx, suby, filter, 
                             in_mosaic, count(filter)
                FROM mosaic_tiles_exposures t, exposure_files e
                WHERE t.expid = e.eid
                AND tile={tile} AND subx={subx} AND suby={suby}
                GROUP BY tile, subx, suby, filter, in_mosaic
                ORDER BY (filter, in_mosaic)
                """)
    return resp


def reset_tiles_in_assoc(assoc):
    """
    Set in_mosaic status to all subtiles overlapping with an assoc
    """
    res = db.execute(f"""UPDATE mosaic_tiles_exposures
     SET in_mosaic = 0
     FROM (select tile, subx, suby
      from mosaic_tiles_exposures ti, exposure_files e
      where ti.expid = e.eid AND e.assoc = '{assoc}'
      group by tile, subx, suby) subt
     WHERE mosaic_tiles_exposures.tile = subt.tile 
          AND mosaic_tiles_exposures.subx = subt.subx
          AND mosaic_tiles_exposures.suby = subt.suby
    """)


def get_tiles_containing_point(point=(150.24727,2.04512), radius=0.01):
    """
    reset in_mosaic status for subtiles overlapping with a point
    """
    
    circle = f"circle '<({point[0]}, {point[1]}), {radius}>'"
    
    res = db.SQL(f"""SELECT tile, subx, suby, filter, count(filter) as nexp
      from mosaic_tiles_exposures ti, exposure_files e
      where ti.expid = e.eid
      AND polygon(e.footprint) && polygon({circle})    
      group by tile, subx, suby, filter
      """)

    return res


def reset_tiles_containing_point(point=(150.24727,2.04512), radius=0.01):
    """
    reset in_mosaic status for subtiles overlapping with a point
    """
    
    circle = f"circle '<({point[0]}, {point[1]}), {radius}>'"
    
    res = db.execute(f"""UPDATE mosaic_tiles_exposures
     SET in_mosaic = 0
     FROM (select tile, subx, suby
      from mosaic_tiles_exposures ti, exposure_files e
      where ti.expid = e.eid
      AND polygon(e.footprint) && polygon({circle})    
      group by tile, subx, suby) subt
      WHERE mosaic_tiles_exposures.tile = subt.tile 
           AND mosaic_tiles_exposures.subx = subt.subx
           AND mosaic_tiles_exposures.suby = subt.suby      
      """)

    return res


def delete_empty_exposures():
    """
    Delete exposures from tiles where exptime = 0
    """
    SQL = f"""SELECT tile, subx, suby, filter, in_mosaic
                FROM mosaic_tiles_exposures t, exposure_files e
                WHERE t.expid = e.eid
                AND e.exptime < 1.
                """
    res = db.SQL(SQL)
    
    if len(res) > 0:
        SQL = f"""DELETE
                    FROM mosaic_tiles_exposures t
                    USING exposure_files e
                    WHERE t.expid = e.eid
                    AND e.exptime < 1.
                    """
    
        db.execute(SQL)


def send_all_tiles():
    
    import time
    import os
    import numpy as np
    import matplotlib.pyplot as plt
    
    import mastquery
    
    from grizli.aws.tile_mosaic import (drizzle_tile_subregion, reset_locked,
                      get_lambda_client, send_event_lambda, count_locked, tile_subregion_wcs)
    
    from grizli.aws import db
    from grizli import utils
    
    tiles = []
    
    client = get_lambda_client()
    
    nt0 = len(tiles)
    
    progs = f'AND tile != 1183 AND tile != 1392'
    progs = f'AND tile != 1183'
    progs = ''
    
    # Get with ra, dec
    if 0:
        tiles = db.SQL(f"""SELECT tile, subx, suby, count(subx),
                           count(distinct(filter)) as nfilt
                    FROM mosaic_tiles_exposures t, exposure_files e
                    WHERE t.expid = e.eid
                    {progs}
                    AND filter LIKE '%%CLEAR'
                    GROUP BY tile, subx, suby
                    ORDER BY count(subx) ASC
                    """)
        
        tiles = db.SQL(f"""SELECT tile, subx, suby, count(filter), filter,
                           SUM(e.exptime) as exptime
                    FROM mosaic_tiles_exposures t, exposure_files e
                    WHERE t.expid = e.eid AND in_mosaic >= 0
                    {progs}
                    AND e.instrume in ('NIRCAM','NIRISS','MIRI')
                    AND filter < 'G0'
                    GROUP BY tile, subx, suby, filter
                    ORDER BY count(subx) ASC
                    """)
        
        tiles['ra'] = 0.
        tiles['dec'] = 0.
        tiles['footprint'] = [np.zeros((4,2))]*len(tiles)
        
        keys = np.array([f'{t} {sx} {sy}'
                         for t, sx, sy in zip(tiles['tile'], tiles['subx'], tiles['suby'])])
                         
        un = utils.Unique(keys, verbose=False)
        for v in tqdm(un.values):
            tile, subx, suby = np.asarray(v.split(),dtype=int)
            _wcs = tile_subregion_wcs(tile, subx, suby)
            ra, dec = _wcs.calc_footprint().mean(axis=0)
            tiles['ra'][un[v]] = ra
            tiles['dec'][un[v]] = dec
            tiles['footprint'][un[v]] = _wcs.calc_footprint()
            
        # Better exposure map
        point = utils.SRegion('circle(53.1, -27.8, 0.2)', wrap=False)
        filt = 'F210M-CLEAR'

        point = utils.SRegion('circle(3.5, -30.8, 2)', wrap=False)
        filt = 'F444W-CLEAR'
        
        coo = np.array([tiles['ra'], tiles['dec']]).T
        in_point = point.path[0].contains_points(coo)
        with_filt = in_point & (tiles['filter'] == filt)
        
        cosd = np.cos(np.median(tiles['dec'][in_point])/180*np.pi)
        
        #plt.plot(*point.xy[0].T, alpha=0.1)
        
        dx = tiles['ra'][in_point].max() - tiles['ra'][in_point].min()
        dy = tiles['dec'][in_point].max() - tiles['dec'][in_point].min()
        aspect = dy/(dx*cosd)
        
        fig, ax = plt.subplots(1,1,figsize=(5,5*aspect))
        
        ax.set_xlim(tiles['ra'][in_point].max()+0.2*dx, 
                    tiles['ra'][in_point].min()-0.2*dx)
        ax.set_ylim(tiles['dec'][in_point].min()-0.2*dy, 
                    tiles['dec'][in_point].max()+0.2*dy)
        
        ax.set_aspect(1./cosd)
        ax.grid()
        ax.set_title(filt)
        
        for f, c in zip(tiles['footprint'][with_filt], tiles['exptime'][with_filt]):
            sr = utils.SRegion(f)
            ic = np.interp(np.log10(c/3600.), [np.log10(100./3600), np.log10(20)],
                           [0, 1], left=0, right=1.)
            for p in sr.get_patch(alpha=0.9, zorder=1000,
                                  fc=plt.cm.magma_r(ic), ec='None'):
                ax.add_patch(p)
                
        _ = mastquery.overlaps.draw_axis_labels(ax, nlabel=3)
        fig.tight_layout(pad=1.0)
        
    ### Reset EGS JWST
    if 0:
        res = db.SQL(f"""select tile, subx, suby, in_mosaic, e.filter
          from mosaic_tiles_exposures ti, exposure_files e
          where ti.tile = 2430 AND ti.expid = e.eid
          AND e.instrume in ('NIRCAM','MIRI')
          AND e.file like 'jw01345%%'
          """)

        res = db.execute(f"""UPDATE mosaic_tiles_exposures
         SET in_mosaic = 0
         FROM (select tile, subx, suby, expid
          from mosaic_tiles_exposures ti, exposure_files e
          where ti.expid = e.eid
          AND e.instrume in ('NIRCAM','MIRI')
          AND e.file like 'jw01345%%' AND e.filter in ('F115W-CLEAR')
               ) subt
      WHERE mosaic_tiles_exposures.tile = subt.tile 
           AND mosaic_tiles_exposures.subx = subt.subx
           AND mosaic_tiles_exposures.suby = subt.suby 
           AND mosaic_tiles_exposures.expid = subt.expid
          """)
        
        
    tiles = db.SQL(f"""SELECT tile, subx, suby, filter, count(filter) as count,
                min(substr(e.file,1,7)) as file0, max(substr(e.file,1,7)) as file1
                FROM mosaic_tiles_exposures t, exposure_files e
                WHERE t.expid = e.eid AND in_mosaic = 0
                {progs}
                AND filter < 'G0'
                GROUP BY tile, subx, suby, filter
                ORDER BY filter ASC
                """)
    
    if 1:
    
        # Skip all optical
        skip = (tiles['filter'] > 'F18') #& (tiles['count'] > 100)
        skip |= (tiles['filter'] < 'F18') & (tiles['count'] > 100)
        timeout = 60
    
        skip = (tiles['filter'] > 'F18') & (tiles['count'] > 100)
        skip |= (tiles['filter'] < 'F18') & (tiles['count'] > 300)
        timeout = 60
        
        tiles = tiles[~skip]
    else:
        # Randomize order for running locally
        ix = np.argsort(np.random.rand(len(tiles)))
        tiles = tiles[ix]
        timeout = 60
        
    nt1 = len(tiles)
    print(nt1, nt0-nt1)
    
    NMAX = len(tiles)

    istart = i = -1
    
    max_locked = 800
    
    step = max_locked - count_locked()[0]
    
    while i < NMAX-1:
        i+=1 
        # if tiles['tile'][i] == 1183:
        #     continue
        
        if i-istart == step:
            istart = i
            print(f'\n ############### \n {time.ctime()}: Pause for {timeout} s  / {step} run previously')
            time.sleep(timeout)
            
            step = np.maximum(max_locked - count_locked()[0], 1)
            print(f'{time.ctime()}: Run {step} more \n ############## \n')
                    
        event = dict(tile=int(tiles['tile'][i]), 
                     subx=int(tiles['subx'][i]),
                     suby=int(tiles['suby'][i]),
                     filter=tiles['filter'][i], 
                     exposure_count=int(tiles['count'][i]),
                     counter=i+2, 
                     time=time.ctime())
        
        if 1:
            send_event_lambda(event, client=client, func='grizli-redshift-fit')
        
        else:
            drizzle_tile_subregion(**event, 
                                   s3output=None,
                                   ir_wcs=None, make_figure=False, 
                                   skip_existing=True, verbose=True, 
                                   gzip_output=False, clean_flt=False)
            
            files='tile.{tile:04d}.{subx:03d}.{suby:03d}.{fx}*fits'
            os.system('rm '+ files.format(fx=event['filter'].lower(), 
                                             **event))
        
        # if (i+1) % 300 == 0:
        #     break
            #time.sleep(10)

    ### Run locally
    tiles = db.SQL(f"""SELECT tile, subx, suby, filter, MAX(in_mosaic) as in_mosaic, count(filter)
                FROM mosaic_tiles_exposures t, exposure_files e
                WHERE t.expid = e.eid AND tile != 1183
                GROUP BY tile, subx, suby, filter
                ORDER BY (tile, filter)
                """)
    
    keep = (tiles['filter'] > 'F199') & (tiles['count'] > 100)
    utils.Unique(tiles['in_mosaic'][keep])
    
    tiles = db.SQL(f"""SELECT tile, subx, suby, filter, count(filter)
                FROM mosaic_tiles_exposures t, exposure_files e
                WHERE t.expid = e.eid AND in_mosaic = 0 AND tile != 1183
                GROUP BY tile, subx, suby, filter
                ORDER BY (tile, filter)
                """)

    skip = (tiles['filter'] > 'F1') & (tiles['count'] < 100)
    skip |= (tiles['filter'] < 'F1') & (tiles['count'] > 300)

    tiles = tiles[~skip]
    
    NMAX = len(tiles)

    istart = i = -1
    
    tile_filt = (0, 0)
    
    while i < NMAX:
        i+=1 
        if tiles['tile'][i] == 1183:
            continue
        
        event = dict(tile=int(tiles['tile'][i]), 
                     subx=int(tiles['subx'][i]),
                     suby=int(tiles['suby'][i]),
                     filter=tiles['filter'][i], 
                     counter=i, 
                     time=time.ctime())
        
        if (event['tile'], event['filter']) != tile_filt:
            print('Next filter, remove flc.fits')
            os.system('rm *flc.fits')
            tile_filt = (event['tile'], event['filter'])
            
        if 0:
            send_event_lambda(event, client=client, func='grizli-mosaic-tile')
        
        else:
            status = drizzle_tile_subregion(**event, 
                                   s3output=None,
                                   ir_wcs=None, make_figure=False, 
                                   clean_flt=False,
                                   skip_existing=True, verbose=True, 
                                   gzip_output=False)
            
            files='tile.{tile:04d}.{subx:03d}.{suby:03d}.{fx}*fits'
            os.system('rm '+ files.format(fx=event['filter'].lower(), 
                                             **event))
    
    
def get_tile_status(tile, subx, suby, filter):
    """
    Is a tile "locked" with all exposures set with in_mosaic = 9?
    """

    exp = db.SQL(f"""SELECT dataset, extension, assoc, filter, 
                            exptime, footprint, in_mosaic, detector
                      FROM mosaic_tiles_exposures t, exposure_files e
                      WHERE t.expid = e.eid
                      AND filter='{filter}' AND tile={tile}
                      AND subx={subx} AND suby={suby}""")
    
    if len(exp) == 0:
        status = 'empty'
    elif (exp['in_mosaic'] == 9).sum() == len(exp):
        status = 'locked'
    elif (exp['in_mosaic'] == 1).sum() == len(exp):
        status = 'completed'
    else:
        status = 'go'
    
    return exp, status


def drizzle_tile_subregion(tile=2530, subx=522, suby=461, filter='F160W', s3output=None, ir_wcs=None, make_figure=False, skip_existing=True, verbose=True, gzip_output=False, **kwargs):
    """
    Drizzle a subtile
    
    Parameters
    ----------
    tile, subx, suby : int
        Identifiers of subtile
    
    filter : str
        Filter bandpass
    
    s3output : str
        Output S3 path, defaults to 
        ``s3://grizli-mosaic-tiles/Tiles/{tile}/``
    
    ir_wcs : `~astropy.wcs.WCS`
        Override subtile WCS
    
    skip_existing : bool
        Skip if output already exists or `in_mosaic` is 0 or 9 in the 
        database
    
    gzip_output : bool
        Gzip the drizzle products
    
    kwargs : dict
        Arguments passed through to `grizli.aws.visit_processor.cutout_mosaic`
        
    Returns
    -------
    status : str
        - ``skip completed`` = `tile.subx.suby.filter` has `in_mosaic = 1` 
           in database
        - ``skip locked`` = `tile.subx.suby.filter` has `in_mosaic = 9` 
           in database
        - ``skip empty`` = no exposures found for `tile.subx.suby.filter`
        - ``skip local`` = `tile.subx.suby.filter` found in local directory
        - ``tile.{tile}.{subx}.{suby}.{filter}`` = rootname of created file
        
    """
    import os
    import astropy.table
    import astropy.units as u
    import astropy.wcs as pywcs
    import astropy.io.fits as pyfits
    
    import numpy as np
    from grizli import utils
    from grizli.aws import visit_processor
    
    exp, status = get_tile_status(tile, subx, suby, filter) 
    
    root = f'tile.{tile:04d}.{subx:03d}.{suby:03d}'
    
    if status in ['empty']:
        if verbose:
            print(f'{root} {filter} ! No exposures found')
            
        return 'skip empty'
    
    elif status in ['locked']:
        print(f'{root} {filter} tile locked')
        return 'skip locked'
    
    elif (status in ['completed']) & (skip_existing):
        print(f'{root} {filter} tile already completed')
        return 'skip completed'
        
    sci_file = f'{root}.{filter.lower()}_drz_sci.fits'
    if skip_existing & os.path.exists(sci_file):
        print(f'Skip file {sci_file}')
        return 'skip local'
    
    # Lock
    db.execute(f"""UPDATE mosaic_tiles_exposures t
          SET in_mosaic = 9
          FROM exposure_files w
          WHERE t.expid = w.eid
          AND w.filter='{filter}' AND tile={tile}
          AND subx={subx} AND suby={suby}""")
        
    if ir_wcs is None:
        ir_wcs = tile_subregion_wcs(tile, subx, suby)
    
    if verbose:
        print(f'{root} {filter} {len(exp)}')
        
    if s3output is None:
        s3output = f's3://grizli-mosaic-tiles/Tiles/{tile}/'

    try:
        visit_processor.cutout_mosaic(
            rootname=root,
            product='{rootname}.{f}',
            ir_wcs=ir_wcs,
            res=exp,
            s3output=s3output,
            make_figure=make_figure,
            skip_existing=skip_existing,
            gzip_output=gzip_output,
            **kwargs
        )

        # Update subtile status
        db.execute(f"""UPDATE mosaic_tiles_exposures t
              SET in_mosaic = 1
              FROM exposure_files w
              WHERE t.expid = w.eid
              AND w.filter='{filter}' AND tile={tile}
              AND subx={subx} AND suby={suby}
               """)

    except TypeError:
        db.execute(f"""UPDATE mosaic_tiles_exposures t
              SET in_mosaic = 8
              FROM exposure_files w
              WHERE t.expid = w.eid
              AND w.filter='{filter}' AND tile={tile}
              AND subx={subx} AND suby={suby}
               """)

    status = '{root}.{f}'
    return status


def query_cutout(output='mos-{tile}-{filter}_{drz}', ra=189.0243001, dec=62.1966953, size=10, filters=['F160W'], theta=0, make_figure=False, make_mosaics=False, all_tiles=True, **kwargs):
    """
    """
    from tqdm import tqdm
    import matplotlib.pyplot as plt
    from grizli import utils
    from grizli.aws import tile_mosaic, db
    
    cosd = np.cos(dec/180*np.pi)
    rc = size/3600*np.sqrt(2)
    rtile = np.sqrt(2)*128*PIXEL_SCALE/3600
    
    SQL = f"""SELECT tile, subx, suby, subra, subdec, filter
            FROM mosaic_tiles_exposures t, exposure_files e
            WHERE in_mosaic = 1
            AND t.expid = e.eid 
            AND ('((' || (subra - {ra})*{cosd} || 
                    ', ' || subdec - {dec} || '),
                    {rtile})')::circle
                && ('((0,0),{rc})')::circle
            GROUP BY tile, subx, suby, subra, subdec, filter"""
    
    res = db.SQL(SQL) 
    if len(res) == 0:
        msg = f'Nothing found for ({ra:.6f}, {dec:.6f}, {size}")'
        return 'empty nothing found', None
        
    if filters is None:
        filters = np.unique(res['filter']).tolist()
    else:
        keep = utils.column_values_in_list(res['filter'], filters)
        res = res[keep]
    
    if len(res) == 0:
        msg = f'Nothing found for ({ra:.6f}, {dec:.6f}, {size}") {filters}'
        return 'empty filter', None
        
    if make_figure:
        fig, ax = plt.subplots(1,1,figsize=(8,8))
        ax.scatter(res['subra'], res['subdec'], marker='x', alpha=0.)
    
    oh, ow = utils.make_wcsheader(ra=ra, dec=dec, size=size*2, 
                                  pixscale=PIXEL_SCALE,
                                  theta=theta)
    
    sh = utils.SRegion(ow)
    shu = sh.union().shapely
    
    if make_figure:
        ax.add_patch(sh.get_patch(alpha=0.2, color='r')[0])
    
    res['keep'] = False
                           
    iters = zip(res['tile'], res['subx'], res['suby'])
    
    keys = [res['tile'][i]*1e6+res['subx'][i]*1000+res['suby'][i]
            for i in range(len(res))]
    
    un = utils.Unique(keys, verbose=False)        
    for k in un.values:
        unk = un[k]
        rk = res[unk][0]
        tw = tile_subregion_wcs(rk['tile'], rk['subx'], rk['suby'])
        sr = utils.SRegion(tw)
        isect = sr.shapely[0].intersects(shu)
        res['keep'][unk] = isect
        
        if make_figure:
            if isect:
                ec = 'b'
                fc = 'b'
                alpha=0.2
            else:
                ec = '0.8'
                fc = 'None'
                alpha=0.3
                
            ax.add_patch(sr.get_patch(alpha=alpha, fc=fc, ec=ec, zorder=-10)[0])
    
    if res['keep'].sum() == 0:
        msg = f'Nothing found for ({ra:.6f}, {dec:.6f}, {size}") {filters}'
        return 'empty filter', None
    
    res = res[res['keep']]
    
    if make_mosaics:
        make_mosaic_from_table(res, output=output, all_tiles=all_tiles,
                               **kwargs)
        
    return 'ok', res


def make_mosaic_from_table(tab, output='mos-{tile}-{filter}_{drz}', clean_subtiles=False, send_to_s3=False, all_tiles=True, **kwargs):
    """
    """
    from grizli import utils
    un = utils.Unique(tab['tile'], verbose=False)
    if all_tiles:
        tiles = un.values
    else:
        tiles = [un.values[np.argmin(un.count)]]
        
    for t in tiles:
        
        unt = tab[un[t]]
        
        xmi = unt['subx'].min()
        ymi = unt['suby'].min()
        xma = unt['subx'].max()
        yma = unt['suby'].max()
        
        filts = [f.lower() for f in np.unique(unt['filter'])]
        
        ll = (xmi, ymi)
        ur = (xma, yma)
        
        for filt in filts:
            build_mosaic_from_subregions(root=output, 
                                         tile=t, files=None, filter=filt, 
                                         ll=ll, ur=ur,
                                         clean_subtiles=clean_subtiles, 
                                         send_to_s3=send_to_s3)
        
        if 0:
            from grizli.pipeline import auto_script
            ds9 = None
            auto_script.field_rgb(root=f'mos-{t}',  HOME_PATH=None, scl=1, ds9=ds9)


def build_mosaic_from_subregions(root='mos-{tile}-{filter}_{drz}', tile=2530, files=None, filter='f140w', ll=None, ur=None, clean_subtiles=False, send_to_s3=False, make_weight=True):
    """
    TBD
    """
    import os
    from tqdm import tqdm
    import glob
    import numpy as np
    import astropy.io.fits as pyfits
    import astropy.wcs as pywcs
    from grizli import utils
    
    if 0:
        for filt in ['f105w','f140w','f160w']:
            build_mosaic_from_subregions(tile=2530, files=None, filter=filt)
            
    #tile = 2530
    #filter = 'f140w'
    
    if files is None:
        files = glob.glob(f'tile.{tile:04d}.*_dr?_sci.fits')
        files.sort()
        
    if len(files) > 0:
        tx = np.array([int(f.split('.')[2]) for f in files])
        ty = np.array([int(f.split('.')[3]) for f in files])
    
        txm, tym = tx.min(), ty.min()
        
    if ll is None:
        ll = [txm, tym]
    
    if ur is None:
        ur = [tx.max(), ty.max()]
           
    nx = ur[0] - ll[0] + 1
    ny = ur[1] - ll[1] + 1
    
    llw = tile_subregion_wcs(tile, ll[0], ll[1])
    
    MIRI_FILTERS = ['f560w','f770w','f1000w','f1280W',
                    'f1500w','f1800w', 'f2100w','f2550w']
                                         
    if ('clear' in filter):
        npix = 512
        llw = utils.half_pixel_scale(llw)
        drz = 'drc'
    elif (filter in MIRI_FILTERS):
        npix = 256
        drz = 'drz'        
    elif (filter > 'f199') & (filter not in ['g102','g141']):
        npix = 512
        llw = utils.half_pixel_scale(llw)
        drz = 'drc'
    else:
        npix = 256
        drz = 'drz'
        
    img = np.zeros((ny*npix, nx*npix), dtype=np.float32)
    if make_weight:
        imgw = np.zeros_like(img)
        
    llh = utils.to_header(llw)
    llh['NAXIS1'] = nx*npix
    llh['NAXIS2'] = ny*npix
    llh['FILTER'] = filter.upper()
    
    ###### and check difference between opt / ir
    ###### and change drizzle params!
    
    exposures = []
    
    llh['EXPSTART'] = 1e10
    llh['EXPEND'] = 0
    
    im = None
    
    for xi in range(ll[0], ur[0]+1):
        for yi in range(ll[1], ur[1]+1):
            file = f'tile.{tile:04d}.{xi:03d}.{yi:03d}'
            file += f'.{filter}_{drz}_sci.fits'
            
            s3 = f's3://grizli-mosaic-tiles/Tiles/{tile}/'
            db.download_s3_file(s3+file, overwrite=False, verbose=False)
            
            if not os.path.exists(file):
                #print('Skip', file)
                continue
        
            print(file)
            im = pyfits.open(file)
                            
            if llh['EXPSTART'] > im[0].header['EXPSTART']:
                llh['EXPSTART'] = im[0].header['EXPSTART']
            
            if llh['EXPEND'] < im[0].header['EXPEND']:
                llh['EXPEND'] = im[0].header['EXPEND']
                
            for j in range(im[0].header['NDRIZIM']):
                exp = im[0].header[f'FLT{j+1:05d}']
                if exp not in exposures:
                    exposures.append(exp)
                    
            slx = slice((xi-ll[0])*npix, (xi-ll[0]+1)*npix)
            sly = slice((yi-ll[1])*npix, (yi-ll[1]+1)*npix)
            img[sly, slx] += im[0].data
            
            if make_weight:
                wfile = file.replace('_sci','_wht')
                db.download_s3_file(s3+wfile, overwrite=False, verbose=False)
                if os.path.exists(wfile):
                    imw = pyfits.open(wfile)
                    imgw[sly, slx] += imw[0].data                                    
            else:
                wfile = None
                
            if clean_subtiles:
                os.remove(file)
                if wfile is not None:
                    os.remove(wfile)
                    
    llh['NDRIZIM'] = len(exposures)
    for j, exp in enumerate(exposures):
        llh[f'FLT{j+1:05d}'] = exp
    
    if im is not None:    
        for k in im[0].header:
            if (k not in llh) & (k not in ['SIMPLE','BITPIX','DATE-OBS','TIME-OBS']):
                llh[k] = im[0].header[k]
                #print(k, im[0].header[k])    
    
    # Empty
    if 'PHOTFLAM' not in llh:
        llh['PHOTFLAM'] = 0.
        llh['PHOTFNU'] = 0.
        llh['PHOTPLAM'] = 1.
            
    outfile = root.format(tile=tile, filter=filter, drz=drz) + '_sci.fits'
    pyfits.writeto(outfile, data=img, 
                   header=llh, overwrite=True)
    
    if make_weight:
        wfile = root.format(tile=tile, filter=filter, drz=drz) + '_wht.fits'
        pyfits.writeto(wfile, data=imgw, header=llh, overwrite=True)
    else:
        wfile = None
        
    if send_to_s3:
        db.upload_file(outfile, 'grizli-v2', object_name='Scratch/'+outfile)
        if make_weight:
            db.upload_file(wfile, 'grizli-v2', object_name='Scratch/'+wfile)
    
    return outfile, wfile
            
    #os.system(f'aws s3 cp {root}.{tile:04d}-{filter}_drz_sci.fits s3://grizli-v2/Scratch/')
    
    



import unittest

import numpy as np

from grizli import utils

from grizli.utils_numba import interp as interp_numba
from grizli.utils_numba import disperse as disperse_numba

# Used to test both utils_numba and utils_c
DISPERSE_MODULES = [disperse_numba]
INTERP_MODULES = [interp_numba]


def test_cinterp():
    """
    Simple interpolation
    """
    xarr = np.array([0.0, 1.0, 2.0])
    yarr = np.array([0.0, 1.0, 0.0])
    for base_module in INTERP_MODULES:
        result = base_module.interp_c(np.array([0.5]), xarr, yarr)
        assert np.allclose(result, 0.5)


def test_cinterp_conserve():
    """
    Linear interpolation conserving the integral
    """
    xarr = np.arange(1.0, 3.0, 0.0001)
    yarr = (np.abs(xarr - 2.0) <= 0.1) * 1.0

    np.random.seed(1)
    xlr = np.random.rand(10) * 2 + 1
    xlr.sort()

    for base_module in INTERP_MODULES:
        ylr = base_module.interp_conserve_c(xlr, xarr, yarr)
        assert np.allclose(np.trapz(ylr, xlr), np.trapz(yarr, xarr))

    # if len(INTERP_MODULES) > 1:
    #     ylr_slow = interp.interp_conserve(xlr, xarr, yarr)
    #     assert np.allclose(np.trapz(ylr_slow, xlr), np.trapz(yarr, xarr))


def test_disperse_c():
    """
    Grism dispersion function
    """
    for disperse_module in DISPERSE_MODULES:
        # Array shapes from a NIRCam F444W example
        sh = (128, 128)
        sh_beam = (128, 1544)
        x0 = np.array([63, 63])

        modelf = np.zeros(sh_beam, dtype=np.float32).flatten()
        seg = np.zeros(sh, dtype=np.float32)

        id = 1.0

        yy, xx = np.indices(sh)
        R = np.sqrt((xx - 62) ** 2 + (yy - 62) ** 2)
        Rmax = 10
        seg += R < Rmax

        nx = sh_beam[1] - sh[0]
        dxpix = np.arange(nx, dtype=int) + sh[0] // 2 - 1
        ytrace_beam = (np.arange(nx) - nx / 2) * 0.5 / 200
        yfrac_beam = ytrace_beam - np.floor(ytrace_beam)
        dyc = (ytrace_beam + 20).astype(int) - 20 + 1
        idx = np.arange(modelf.size, dtype=np.int64).reshape(sh_beam)
        flat_index = idx[dyc + x0[0], dxpix]
        sens_curve = yfrac_beam**0

        ########
        # single pixel
        thumb = seg * 0.0
        thumb[63, 63] = 1.0

        modelf *= 0.0

        status = disperse_module.disperse_grism_object(
            thumb,
            seg,
            np.float32(id),
            flat_index,
            yfrac_beam.astype(np.float64),
            sens_curve.astype(np.float64),
            modelf,
            x0,
            np.array(sh, dtype=np.int64),
            x0,
            np.array(sh_beam, dtype=np.int64),
        )

        model2d = modelf.reshape(sh_beam)
        model_trace = model2d[:, x0[1] : -(x0[1] + 2)]

        # "left" edge of 2D model is zero
        assert np.allclose(model2d[:, : x0[1]], 0.0)

        npix = (model_trace > 0).sum(axis=0)
        assert npix.max() >= 1
        assert npix.max() <= 2

        # Pixels across the trace should sum to one
        assert np.allclose(model_trace.sum(axis=0), 1.0)

        ########
        # A "source" with flux across the segment
        thumb = seg * 1.0

        modelf *= 0.0

        status = disperse_module.disperse_grism_object(
            thumb,
            seg,
            np.float32(id),
            flat_index,
            yfrac_beam.astype(np.float64),
            sens_curve.astype(np.float64),
            modelf,
            x0,
            np.array(sh, dtype=np.int64),
            x0,
            np.array(sh_beam, dtype=np.int64),
        )

        model2d = modelf.reshape(sh_beam)
        model_trace = model2d[:, x0[1] + Rmax : -(x0[1] + 2 + Rmax)]

        # Each location along the spectrum should see the "whole" thumb
        assert np.allclose(model_trace.sum(axis=0), thumb.sum())

        #######
        # Unresolved "emission line" in a single spectral bin should give the thumbnail
        line_spectrum = sens_curve * 0
        line_spectrum[sh_beam[1] // 2] = 1
        modelf *= 0.0

        status = disperse_module.disperse_grism_object(
            thumb,
            seg,
            np.float32(id),
            flat_index,
            yfrac_beam.astype(np.float64),
            line_spectrum.astype(np.float64),
            modelf,
            x0,
            np.array(sh, dtype=np.int64),
            x0,
            np.array(sh_beam, dtype=np.int64),
        )

        model2d = modelf.reshape(sh_beam)
        model_trace = model2d[:, x0[1] + Rmax : -(x0[1] + 2 + Rmax)]

        assert model_trace.sum() == thumb.sum()


def test_segmentation_limits():
    """
    Accelerator function to calculate segmentation cutouts
    """
    sh = (128, 128)
    x0 = (32, 38)
    
    yp, xp = np.indices(sh)
    
    flam = np.sqrt((xp - x0[0]) ** 2 + (yp - x0[1]) ** 2).astype(np.float32)
    
    Rsize = 5
    segm = (flam < Rsize).astype(np.float32)

    total_flux = flam[segm > 0].sum()

    for disperse_module in DISPERSE_MODULES:

        _ = disperse_module.compute_segmentation_limits(
            segm, 1.0, flam, np.array(sh, dtype=int)
        )

        imin, imax, ic, jmin, jmax, jc, area, tot_i = _

        assert imin == (x0[1] - (Rsize - 1))
        assert imax == (x0[1] + (Rsize - 1))
        assert jmin == (x0[0] - (Rsize - 1))
        assert jmax == (x0[0] + (Rsize - 1))
        assert area == int(segm.sum())
        assert np.allclose(tot_i, total_flux, rtol=1.0e-5)

        _ = disperse_module.compute_segmentation_limits(
            segm, 2.0, flam, np.array(sh, dtype=int)
        )

        imin, imax, ic, jmin, jmax, jc, area, tot_i = _
        assert area == 0
        assert np.allclose(tot_i, -99.0, rtol=1.0e-5)



import unittest

import os

import numpy as np

from grizli import grismconf as grizliconf
from grizli import GRIZLI_PATH


def test_transform():
    """
    Test JWST transforms
    """
    import astropy.io.fits as pyfits
    
    for instrument in ['NIRISS','NIRCAM','WFC3']:
        for grism in 'RC':
            for module in 'AB':
                
                tr = grizliconf.JwstDispersionTransform(instrument=instrument,
                                            grism=grism, module=module)
                
                #print(instrument, grism, module, tr.forward(1024, 1024))
                assert(np.allclose(tr.forward(1024.5, 1024.5), 1024.5))
                
                # Forward, Reverse
                x0 = np.array([125., 300])
                to = tr.forward(*x0)
                fro = np.squeeze(tr.reverse(*to))
                
                assert(np.allclose(x0-fro, 0.))

    # From header
    nis = pyfits.Header()
    nis['INSTRUME'] = 'NIRISS'

    rot90 = {'GR150C':2,  # 180 degrees
             'GR150R':3,   # 270 degrees CW
             }

    for gr in rot90:
        nis['FILTER'] = gr
        nis['PUPIL'] = 'F150W'

        tr = grizliconf.JwstDispersionTransform(header=nis)
        assert(tr.instrument == nis['INSTRUME'])
        assert(tr.grism == nis['FILTER'])
        assert(tr.rot90 == rot90[gr])


def test_read():
    
    CONF_PATH = os.path.join(GRIZLI_PATH, 'CONF')
    
    wfc3_file = os.path.join(CONF_PATH, 'G141.F140W.V4.32.conf')
    if os.path.exists(wfc3_file):
        conf = grizliconf.aXeConf(wfc3_file)
        
    try:
        import grismconf
        has_grismconf = True
    except ImportError:
        has_grismconf = False
    
    wfc3_gc = os.path.join(CONF_PATH, 'GRISM_WFC3/IR/G141.conf')
    if os.path.exists(wfc3_gc) & has_grismconf:
        conf = grizliconf.TransformGrismconf(conf_file=wfc3_gc)
        


"""
Numba-accelerated functions for filtering with `scipy.LowLevelallable`

Inspired by the beautifully simply example from Juan Nunez-Iglesias
https://ilovesymposia.com/2017/03/12/scipys-new-lowlevelcallable-is-a-game-changer/

>>> import numpy as np
>>> import scipy.ndimage as nd
>>> import grizli.nbutils
>>> arr = np.random.normal(size=1024)
>>> footprint = np.ones(51, dtype=int)
>>> # Slow with numpy overheads
>>> np_filtered = nd.generic_filter(arr, np.nanmedian, footprint=footprint))
>>> # Much faster!
>>> filtered = nd.generic_filter(arr, grizli.nbutils.nanmedian, footprint=footprint))

"""

import numpy as np

from numba import cfunc, carray
from numba.types import intc, intp, float64, voidptr
from numba.types import CPointer

from scipy import LowLevelCallable


@cfunc(intc(CPointer(float64), intp, CPointer(float64), voidptr))
def nb_nanmin(values_ptr, len_values, result, data):
    """
    Manual minimum
    """
    values = carray(values_ptr, (len_values,), dtype=float64)
    result[0] = np.inf
    for v in values:
        if (v < result[0]) & ~np.isnan(v):
            result[0] = v
    return 1


@cfunc(intc(CPointer(float64), intp, CPointer(float64), voidptr))
def nb_nanmax(values_ptr, len_values, result, data):
    """
    Manual minimum
    """
    values = carray(values_ptr, (len_values,), dtype=float64)
    result[0] = -np.inf
    for v in values:
        if (v > result[0]) & ~np.isnan(v):
            result[0] = v
    return 1


@cfunc(intc(CPointer(float64), intp, CPointer(float64), voidptr))
def nb_npnanmin(values_ptr, len_values, result, data):
    """
    numpy.nanmin
    """
    values = carray(values_ptr, (len_values,), dtype=float64)
    result[0] = np.nanmin(values)
    return 1


@cfunc(intc(CPointer(float64), intp, CPointer(float64), voidptr))
def nb_npnanmedian(values_ptr, len_values, result, data):
    """
    numpy.nanmedian
    """
    values = carray(values_ptr, (len_values,), dtype=float64)
    result[0] = np.nanmedian(values)
    return 1


@cfunc(intc(CPointer(float64), intp, CPointer(float64), voidptr))
def nb_npnanmean(values_ptr, len_values, result, data):
    """
    numpy.nanmean
    """
    values = carray(values_ptr, (len_values,), dtype=float64)
    result[0] = np.nanmean(values)

    return 1


@cfunc(intc(CPointer(float64), intp, CPointer(float64), voidptr))
def nb_npnansum(values_ptr, len_values, result, data):
    """
    numpy.nansum
    """
    values = carray(values_ptr, (len_values,), dtype=float64)
    result[0] = np.nansum(values)

    return 1


# Callables for scipy.ndimage.generic_filter
nanmin = LowLevelCallable(nb_nanmin.ctypes)
nanmax = LowLevelCallable(nb_nanmax.ctypes)
nbnanmin = LowLevelCallable(nb_npnanmin.ctypes)
nanmedian = LowLevelCallable(nb_npnanmedian.ctypes)
nanmean = LowLevelCallable(nb_npnanmean.ctypes)
nansum = LowLevelCallable(nb_npnansum.ctypes)



"""
Script to run all redshift fits in parallel with OpenMPI

Usage:

    if [ -e $GRIZLICODE ]; then
        GRIZLICODE=`python -c "import grizli; import os; print(os.path.dirname(grizli.__file__))"`
    fi

    mpiexec -n 8 python -m mpi4py.futures $GRIZLICODE/grizli/pipeline/run_MPI.py

where "-n 8" indicates running 8 parallel threads.

Needs 'fit_args.py' created by `auto_script.generate_fit_params`.

"""
import time
import os
import glob

import numpy as np

import matplotlib.pyplot as plt

from grizli.fitting import run_all_parallel
from grizli import utils

plt.ioff()
utils.set_warnings()


def find_ids():
    # Find objects that with extarcted spectra and that need to be fit
    all_files = glob.glob('*beams.fits')
    files = []
    for file in all_files:
        if not os.path.exists(file.replace('beams.fits', 'full.fits')):
            files.append(file)

    print('{0} files to fit'.format(len(files)))

    ids = [int(file.split('_')[1].split('.')[0]) for file in files]

    return ids


if __name__ == '__main__':

    from mpi4py.futures import MPIPoolExecutor
    import drizzlepac  # In here for travis

    t1 = time.time()

    ids = find_ids()
    if len(ids) == 0:
        exit()

    with MPIPoolExecutor() as executor:
        res = executor.map(run_all_parallel, ids)
        for ix in res:
            print('  Done, id={0} / status={1}, t={2:.1f}'.format(ix[0], ix[1], ix[2]))

    t2 = time.time()

    print('MPIPool: {0:.1f}'.format(t2-t1))



"""
Scripts to combine FLT exposures at a single orientation / sub-pixel position
to speed up the spectral processing.

2024-09-92 (GBr): hasn't been used / tested for many years
"""
import copy
import glob

import numpy as np

import matplotlib.pyplot as plt

import astropy.wcs as pywcs
import astropy.io.fits as pyfits

from . import utils
from . import GRIZLI_PATH


def combine_flt(
    files=[],
    output="exposures_cmb.fits",
    grow=1,
    add_padding=True,
    pixfrac=0.5,
    kernel="point",
    verbose=True,
    overwrite=True,
    ds9=None,
):
    """
    Drizzle distorted FLT frames to an "interlaced" image

    Parameters
    ----------
    files : list of strings
        Filenames of FLT files to combine

    output : str
        Output filename of the combined file.  Convention elsewhere is to use
        an "_cmb.fits" extension to distinguish from "_flt.fits".

    grow : int
        Factor by which to `grow` the FLT frames to interlaced outputs.  For
        example, `grow=2` results in 2x2 interlacing.

    add_padding : True
        Expand pixel grid to accommodate all dithered exposures.  WCS is
        preserved but "CRPIX" will change.

    pixfrac : float
        Drizzle pixfrac (for kernels other than 'point')

    kernel : {'point', 'square'}
        Drizzle kernel. The 'point' kernel is effectively interlacing and is
        best for preserving the noise properties of the final combined image.
        However, can result in empty pixels given the camera distortions
        depending on the dithering of the input exposures.

    verbose : bool
        Print logging information

    overwrite : bool
        Overwrite existing files

    ds9 : `~grizli.ds9.DS9`
        Display the progress of the script to a DS9 window.

    Returns
    -------
    Creates combined images

    """
    import numpy.linalg
    from stsci.tools import asnutil
    from drizzlepac import astrodrizzle

    # `files` is an ASN filename, not  a list of exposures
    if "_asn.fits" in files:
        asn = asnutil.readASNTable(files)
        files = ["{0}_flt.fits".format(flt) for flt in asn["order"]]
        if output == "combined_flt.fits":
            output = "{0}_cmb.fits".format(asn["output"])

    f0 = pyfits.open(files[0])
    h0 = f0[0].header.copy()
    h0["EXPTIME"] = 0.0
    h0["NFILES"] = (len(files), "Number of combined files")

    out_wcs = pywcs.WCS(f0[1].header, relax=True)
    out_wcs.pscale = utils.get_wcs_pscale(out_wcs)
    # out_wcs.pscale = np.sqrt(out_wcs.wcs.cd[0,0]**2 +
    #                          out_wcs.wcs.cd[1,0]**2)*3600.

    # Compute maximum offset needed for padding
    if add_padding:
        ra0, de0 = out_wcs.all_pix2world([0], [0], 0)

        x0 = np.zeros(len(files))
        y0 = np.zeros(len(files))

        for i, file in enumerate(files):
            hx = pyfits.getheader(file, 0)
            h0["EXPTIME"] += hx["EXPTIME"]
            h0["FILE{0:04d}".format(i)] = (file, "Included file #{0:d}".format(i))

            h = pyfits.getheader(file, 1)
            flt_wcs = pywcs.WCS(h, relax=True)
            x0[i], y0[i] = flt_wcs.all_world2pix(ra0, de0, 0)

        xmax = np.abs(x0).max()
        ymax = np.abs(y0).max()
        padx = 50 * int(np.ceil(xmax / 50.0))
        pady = 50 * int(np.ceil(ymax / 50.0))
        pad = np.maximum(padx, pady) * grow

        if verbose:
            print(
                "Maximum shift (x, y) = ({0:6.1f}, {1:6.1f}), pad={2:d}".format(
                    xmax, ymax, pad
                )
            )
    else:
        pad = 0

    inter_wcs = out_wcs.deepcopy()
    if grow > 1:
        inter_wcs.wcs.cd /= grow
        for i in range(inter_wcs.sip.a_order + 1):
            for j in range(inter_wcs.sip.a_order + 1):
                inter_wcs.sip.a[i, j] /= grow ** (i + j - 1)

        for i in range(inter_wcs.sip.b_order + 1):
            for j in range(inter_wcs.sip.b_order + 1):
                inter_wcs.sip.b[i, j] /= grow ** (i + j - 1)

        if hasattr(inter_wcs, "_naxis1"):
            inter_wcs._naxis1 *= grow
            inter_wcs._naxis2 *= grow
        else:
            for i in range(len(inter_wcs._naxis)):
                inter_wcs._naxis[i] *= grow

        inter_wcs.wcs.crpix *= grow
        inter_wcs.sip.crpix[0] *= grow
        inter_wcs.sip.crpix[1] *= grow

        if grow > 1:
            inter_wcs.wcs.crpix += grow / 2.0
            inter_wcs.sip.crpix[0] += grow / 2.0
            inter_wcs.sip.crpix[1] += grow / 2.0

    if hasattr(inter_wcs, "_naxis1"):
        inter_wcs._naxis1 += pad
        inter_wcs._naxis2 += pad
    else:
        for i in range(len(inter_wcs._naxis)):
            inter_wcs._naxis[i] += pad

    inter_wcs.wcs.crpix += pad
    inter_wcs.sip.crpix[0] += pad
    inter_wcs.sip.crpix[1] += pad

    outh = inter_wcs.to_header(relax=True)
    for key in outh:
        if key.startswith("PC"):
            outh.rename_keyword(key, key.replace("PC", "CD"))

    outh["GROW"] = grow, "Grow factor"
    outh["PAD"] = pad, "Image padding"
    outh["BUNIT"] = h["BUNIT"]

    sh = (1014 * grow + 2 * pad, 1014 * grow + 2 * pad)
    outsci = np.zeros(sh, dtype=np.float32)
    outwht = np.zeros(sh, dtype=np.float32)
    outctx = np.zeros(sh, dtype=np.int32)

    # Pixel area map
    # PAM_im = pyfits.open(os.path.join(os.getenv('iref'), 'ir_wfc3_map.fits'))
    # PAM = PAM_im[1].data

    for i, file in enumerate(files):
        im = pyfits.open(file)

        if verbose:
            print(
                "{0:3d} {1:s} {2:6.1f} {3:6.1f} {4:10.2f}".format(
                    i + 1, file, x0[i], y0[i], im[0].header["EXPTIME"]
                )
            )

        dq = utils.mod_dq_bits(im["DQ"].data, okbits=608, verbose=False)
        wht = 1.0 / im["ERR"].data ** 2
        wht[(im["ERR"].data == 0) | (dq > 0) | (~np.isfinite(wht))] = 0
        wht[im["SCI"].data < -3 * im["ERR"].data] = 0

        wht = np.asarray(wht, dtype=np.float32)

        exp_wcs = pywcs.WCS(im[1].header, relax=True)
        exp_wcs.pscale = utils.get_wcs_pscale(exp_wcs)

        # pf = 0.5
        # import drizzlepac.wcs_functions as dwcs
        # xx = out_wcs.deepcopy()
        # #xx.all_pix2world = xx.wcs_world2pix
        # map = dwcs.WCSMap(exp_wcs, xx)

        astrodrizzle.adrizzle.do_driz(
            im["SCI"].data,
            exp_wcs,
            wht,
            inter_wcs,
            outsci,
            outwht,
            outctx,
            1.0,
            "cps",
            1,
            wcslin_pscale=exp_wcs.pscale,
            uniqid=1,
            pixfrac=pixfrac,
            kernel=kernel,
            fillval=0,
            stepsize=10,
            wcsmap=SIP_WCSMap,
        )

        if ds9 is not None:
            ds9.view(outsci, header=outh)

    # outsci /= out_wcs.pscale**2
    rms = 1 / np.sqrt(outwht)
    mask = (outwht == 0) | (rms > 100)
    rms[mask] = 0
    outsci[mask] = 0.0

    hdu = [pyfits.PrimaryHDU(header=h0)]
    hdu.append(pyfits.ImageHDU(data=outsci / grow ** 2, header=outh, name="SCI"))
    hdu.append(pyfits.ImageHDU(data=rms / grow ** 2, header=outh, name="ERR"))
    hdu.append(pyfits.ImageHDU(data=mask * 1024, header=outh, name="DQ"))

    pyfits.HDUList(hdu).writeto(output, overwrite=overwrite, output_verify="fix")


def combine_visits_and_filters(
    grow=1,
    pixfrac=0.5,
    kernel="point",
    filters=["G102", "G141"],
    skip=None,
    split_visit=False,
    split_quadrants=True,
    overwrite=True,
    ds9=None,
    verbose=True,
):
    """
    Make combined FLT files for all FLT files in the working directory
    separated by targname/visit/filter

    Parameters
    ----------
    grow : int
        Factor by which to `grow` the FLT frames to interlaced outputs.  For
        example, `grow=2` results in 2x2 interlacing.

    pixfrac : float
        Drizzle pixfrac (for kernels other than 'point')

    kernel : {'point', 'square'}
        Drizzle kernel. The 'point' kernel is effectively interlacing and is
        best for preserving the noise properties of the final combined image.
        However, can result in empty pixels given the camera distortions
        depending on the dithering of the input exposures.
    
    filters : list of strings
        Only make products for exposures that use these filters
    
    skip : `slice` or None
        Slice of the overall list of visits to process a subset

    split_visit : bool
        If `True`, then separate by all TARGNAME/visit otherwise group by
        TARGNAME and combine visits.

    split_quadrants : bool
        Split by 2x2 sub-pixel dither positions

    overwrite : bool
        Overwrite existing files

    ds9 : `pyds9.DS9`
        Display the progress of the script to a DS9 window.

    verbose : bool
        Print logging information

    Returns
    -------
    nothing but creates "cmb" combined files
    """
    
    files = glob.glob("i*flt.fits")
    output_list, xx = utils.parse_flt_files(files, uniquename=split_visit)

    if skip is None:
        skip = slice(0, len(output_list))

    for i in range(len(output_list))[skip]:
        key = output_list[i]["product"]
        filter = key.split("-")[-1].upper()
        if filter not in filters:
            continue

        output = "{0}_cmb.fits".format(key)

        if verbose:
            print("\n -- Combine: {0} -- \n".format(output))

        if split_quadrants:
            combine_quadrants(
                files=output_list[i]["files"],
                output=output,
                ref_pixel=[507, 507],
                pixfrac=pixfrac,
                kernel=kernel,
                overwrite=overwrite,
                ds9=ds9,
                verbose=verbose,
            )
        else:
            combine_flt(
                files=output_list[i]["files"],
                output=output,
                grow=grow,
                ds9=ds9,
                verbose=verbose,
                overwrite=overwrite,
                pixfrac=pixfrac,
                kernel=kernel,
            )


def get_shifts(files, ref_pixel=[507, 507]):
    """
    Compute relative pixel shifts based on header WCS

    Parameters
    ----------
    files : list of exposure filenames

    ref_pixel : [int, int] or [float, float]
        Reference pixel for the computed shifts

    Returns
    -------
    h : `~astropy.io.fits.Header`
        Header of the first exposure modified with the total exposure time
        and filenames of the input files in the combination.

    xshift, yshift : array-like
        Computed pixel shifts

    """
    f0 = pyfits.open(files[0])
    h0 = f0[0].header.copy()
    h0["EXPTIME"] = 0.0

    out_wcs = pywcs.WCS(f0[1].header, relax=True)
    out_wcs.pscale = utils.get_wcs_pscale(out_wcs)

    # Offsets
    ra0, de0 = out_wcs.all_pix2world([ref_pixel[0]], [ref_pixel[1]], 0)

    x0 = np.zeros(len(files))
    y0 = np.zeros(len(files))

    for i, file in enumerate(files):
        hx = pyfits.getheader(file, 0)
        h0["EXPTIME"] += hx["EXPTIME"]
        h0["FILE{0:04d}".format(i)] = file, "Included file #{0:d}".format(i)

        h = pyfits.getheader(file, 1)
        flt_wcs = pywcs.WCS(h, relax=True)
        x0[i], y0[i] = flt_wcs.all_world2pix(ra0, de0, 0)

    return h0, x0 - ref_pixel[0], y0 - ref_pixel[1]


def split_pixel_quadrant(dx, dy, figure="quadrants.png", show=False):
    """
    Group offsets by their sub-pixel quadrant

    Parameters
    ----------
    dx, dy : array-like
        Pixel shifts of a list of exposures, for example output from
        `~grizli.combine.get_shifts`.

    figure : str
        If not an empty string, save a diagnostic figure showing the
        derived sub-pixel quadrants.

        .. plot::
            :include-source:

            import matplotlib.pyplot as plt
            import numpy as np
            import grizli.combine

            ### 3D-HST dither pattern
            dx = np.array([0, 10, 6.5, -3.5])
            dy = np.array([0, 3.5, 10, 6.5])

            ### This computes the quadrants and also generates the plot
            quad = grizli.combine.split_pixel_quadrant(dx, dy, show=True)
            print(quad)
            # {0: array([0]), 1: array([2]), 2: array([1]), 3: array([3])}
            plt.show()

    show : bool
        Don't close the generated figure (for online docs)

    Returns
    -------
    quad : dict
        Dictionary with keys of integers specifying each of 4 sub-pixel
        quadrants and entries of array indices based on the input `dx` and
        `dy` arrays.

    """
    from matplotlib.ticker import MultipleLocator

    xq = np.asarray(np.round((dx - np.floor(dx)) * 2), dtype=int) % 2
    yq = np.asarray(np.round((dy - np.floor(dy)) * 2), dtype=int) % 2

    # Test, show sub-pixel centers in a figure
    if figure:
        xf = dx - np.floor(dx)
        yf = dy - np.floor(dy)

        colors = np.array([["r", "g"], ["b", "k"]])

        fig = plt.figure(figsize=[6, 6])
        ax = fig.add_subplot(111)

        ax.scatter(xf, yf, c=colors[xq, yq], marker="o", alpha=0.8)

        box = np.array([-0.25, 0.25])
        for i in range(2):
            for j in range(2):
                ax.fill_between(
                    j * 0.5 + box,
                    i * 0.5 + 0.25,
                    i * 0.5 - 0.25,
                    color=colors[j, i],
                    alpha=0.05,
                )

                ax.fill_between(
                    j * 0.5 + box + 1,
                    i * 0.5 + 0.25 + 1,
                    i * 0.5 - 0.25 + 1,
                    color=colors[j, i],
                    alpha=0.05,
                )

                ax.fill_between(
                    j * 0.5 + box,
                    i * 0.5 + 0.25 + 1,
                    i * 0.5 - 0.25 + 1,
                    color=colors[j, i],
                    alpha=0.05,
                )

                ax.fill_between(
                    j * 0.5 + box + 1,
                    i * 0.5 + 0.25,
                    i * 0.5 - 0.25,
                    color=colors[j, i],
                    alpha=0.05,
                )

        ax.set_xlabel(r"Sub-pixel $x$")
        ax.set_ylabel(r"Sub-pixel $y$")
        ax.set_xlim(-0.1, 1.1)
        ax.set_ylim(-0.1, 1.1)
        majorLocator = MultipleLocator(0.25)
        for axis in [ax.xaxis, ax.yaxis]:
            axis.set_major_locator(majorLocator)

        ax.grid()

        if not plt.rcParams["interactive"]:
            if not show:
                fig.savefig(figure)
                plt.close(fig)

    index = np.arange(len(dx))
    out = {}
    for i in range(4):
        match = xq + 2 * yq == i
        if match.sum() > 0:
            out[i] = index[match]

    return out


# Superceded by `combine_visits_and_filters`
# def combine_figs():
#     """
#     Combine FIGS exposures split by offset pixel quadrant
#     """
#     from stsci.tools import asnutil
#
#     #all_asn = glob.glob('figs-g*-g1*asn.fits')
#     all_asn = []
#     all_asn.extend(glob.glob('g[ns][0-9]*g102*asn.fits'))
#     all_asn.extend(glob.glob('gdn*g102*asn.fits'))
#     grism = 'g102'
#
#     #all_asn = glob.glob('gn-z10*-g1*asn.fits')
#     #all_asn.extend(glob.glob('colfax-*g14*asn.fits'))
#     #all_asn = glob.glob('goodsn-*g14*asn.fits')
#     #grism = 'g141'
#
#     roots = np.unique(['-'.join(asn.split('-')[:2]) for asn in all_asn])
#     for root in roots:
#         all_asn = glob.glob('%s*%s*asn.fits' %(root, grism))
#         angles = np.unique([asn.split('-')[-2] for asn in all_asn])
#         for angle in angles:
#             asn_files = glob.glob('%s*-%s-%s*asn.fits' %(root, angle, grism))
#
#             grism_files = []
#             for file in asn_files:
#                 asn = asnutil.readASNTable(file)
#                 grism_files.extend(['%s_flt.fits' %(flt) for flt in asn['order']])
#
#             print '%s-%s %d' %(root, angle, len(grism_files))
#             combine_quadrants(files=grism_files, output='%s-%s-%s_cmb.fits' %(root, angle, grism))


def combine_quadrants(
    files=[],
    output="images_cmb.fits",
    grow=1,
    pixfrac=0.5,
    kernel="point",
    ref_pixel=[507, 507],
    ds9=None,
    verbose=True,
    overwrite=True,
):
    """
    Wrapper to split a list of exposures based on their shift sub-pixel quadrants

    Parameters are all passed directly to `~grizli.combine.combine_flt` but
    separated by shift quadrant with `~grizli.combine.get_shifts` and
    `~grizli.combine.split_pixel_quadrant`.

    Parameters
    ----------
    files : list
        List of exposure filenames

    output : str
        Basename of the output file.  Must end in "_cmb.fits" because the
        actual output files are derived with the following:

        >>> quad_output = output.replace('_cmb.fits',
                                         '_q{0:d}_cmb.fits'.format(q))

    grow : int
        Factor by which to `grow` the FLT frames to interlaced outputs.  For
        example, `grow=2` results in 2x2 interlacing.

    pixfrac : float
        Drizzle pixfrac (for kernels other than 'point')
    
    kernel : {'point', 'square'}
        Drizzle kernel. The 'point' kernel is effectively interlacing and is
        best for preserving the noise properties of the final combined image.
        However, can result in empty pixels given the camera distortions
        depending on the dithering of the input exposures.

    ref_pixel : [int, int] or [float, float]
        Reference pixel for the computed shifts
    
    ds9 : `~grizli.ds9.DS9`
        Display the progress of the script to a DS9 window.

    verbose : bool
        Print logging information

    overwrite : bool
        Overwrite existing files
    """
    h, dx, dy = get_shifts(files, ref_pixel=ref_pixel)
    out = split_pixel_quadrant(dx, dy, figure=output.replace(".fits", "_quad.png"))

    for q in out:
        print("Quadrant {0:d}, {1:d} files".format(q, len(out[q])))
        quad_output = output.replace("_cmb.fits", "_q{0:d}_cmb.fits".format(q))
        combine_flt(
            files=np.array(files)[out[q]],
            output=quad_output,
            grow=grow,
            pixfrac=pixfrac,
            kernel=kernel,
            ds9=ds9,
            verbose=verbose,
            overwrite=overwrite,
        )


# Default mapping function based on PyWCS


class SIP_WCSMap:
    def __init__(self, input, output, origin=1):
        """
        Sample class to demonstrate how to define a coordinate transformation

        Modified from `drizzlepac.wcs_functions.WCSMap` to use full SIP header
        in the `forward` and `backward` methods. Use this class to drizzle to
        an output distorted WCS, e.g.,

            >>> drizzlepac.astrodrizzle.do_driz(..., wcsmap=SIP_WCSMap)

        Parameters
        ----------
        input : `astropy.wcs.WCS`
            Input WCS object.

        output : `astropy.wcs.WCS`
            Output WCS object.

        origin : int
            Origin of the pixel coordinates (default is 1).
        
        """

        # Verify that we have valid WCS input objects
        self.checkWCS(input, "Input")
        self.checkWCS(output, "Output")

        self.input = input
        self.output = copy.deepcopy(output)
        # self.output = output

        self.origin = origin
        self.shift = None
        self.rot = None
        self.scale = None

    def checkWCS(self, obj, name):
        """
        Check that the input object is a valid PyWCS object.
        
        Parameters
        ----------
        obj : object
            Object to be checked.

        name : str
            Name of the object to be checked.

        """
        try:
            assert isinstance(obj, pywcs.WCS)
        except AssertionError:
            print(
                name + " object needs to be an instance or subclass of a PyWCS object."
            )
            raise

    def forward(self, pixx, pixy):
        """
        Transform the input pixx,pixy positions in the input frame
        to pixel positions in the output frame.

        This method gets passed to the drizzle algorithm.

        Parameters
        ----------
        pixx : array-like
            X pixel positions in the input frame.

        pixy : array-like
            Y pixel positions in the input frame.

        Returns
        -------
        result : tuple
            Tuple of transformed pixel positions in the output frame.

        """
        # This matches WTRAXY results to better than 1e-4 pixels.
        skyx, skyy = self.input.all_pix2world(pixx, pixy, self.origin)
        # result= self.output.wcs_world2pix(skyx,skyy,self.origin)
        result = self.output.all_world2pix(skyx, skyy, self.origin)
        return result

    def backward(self, pixx, pixy):
        """
        Transform pixx,pixy positions from the output frame back onto their
        original positions in the input frame.

        Parameters
        ----------
        pixx : array-like
            X pixel positions in the output frame.
        
        pixy : array-like
            Y pixel positions in the output frame.

        Returns
        -------
        result : tuple
            Tuple of transformed pixel positions in the input frame.

        """
        # skyx,skyy = self.output.wcs_pix2world(pixx,pixy,self.origin)
        skyx, skyy = self.output.all_pix2world(pixx, pixy, self.origin)
        result = self.input.all_world2pix(skyx, skyy, self.origin)
        return result

    def get_pix_ratio(self):
        """
        Return the ratio of plate scales between the input and output WCS.
        This is used to properly distribute the flux in each pixel in 'tdriz'.
        """
        return self.output.pscale / self.input.pscale

    def xy2rd(self, wcs, pixx, pixy):
        """
        Transform input pixel positions into sky positions in the WCS provided.
        
        Parameters
        ----------
        wcs : `astropy.wcs.WCS`
            WCS object representing the coordinate transformation.

        pixx : array-like
            X pixel positions in the input frame.

        pixy : array-like
            Y pixel positions in the input frame.

        Returns
        -------
        result : tuple
            Tuple of transformed sky positions in the output frame.

        """
        return wcs.all_pix2world(pixx, pixy, 1)

    def rd2xy(self, wcs, ra, dec):
        """
        Transform input sky positions into pixel positions in the WCS provided.
        
        Parameters
        ----------
        wcs : `astropy.wcs.WCS`
            WCS object representing the coordinate transformation.

        ra : array-like
            Right ascension positions in degrees.

        dec : array-like
            Declination positions in degrees.

        Returns
        -------
        result : tuple
            Tuple of transformed pixel positions in the output frame.

        """
        return wcs.wcs_world2pix(ra, dec, 1)



"""
Tests for spectra, don't implement now because will fail on
travis where we haven't set up the full GRIZLI path and reference
files
"""
import os
import glob
import unittest

import numpy as np
from grizli import utils, prep, multifit, fitting, GRIZLI_PATH


class FittingTools(unittest.TestCase):
    
    def test_config(self):
        """
        Fetch config files if CONF not found
        """
        new = []
        for subd in ['iref','jref','CONF']:
            conf_path = os.path.join(GRIZLI_PATH, subd)
            if not os.path.exists(conf_path):
                new.append(subd)
                os.mkdir(conf_path)
        
        if 'CONF' in new:
            print(f'Download config and calib files to {conf_path}')
            utils.fetch_default_calibs(get_acs=False)
            utils.fetch_config_files(get_epsf=True)
            files = glob.glob(f'{conf_path}/*')
            print('Files: ', '\n'.join(files))

        assert os.path.exists(os.path.join(conf_path,
                              'G141.F140W.V4.32.conf'))
            
    def test_multibeam(self):
        """
        Can we initialize a multibeam file?
        """
        path = os.path.dirname(utils.__file__)
        print(path)
        beams_file = os.path.join(path, 'tests', 'data', 'j033216m2743_00152.beams.fits')
        mb = multifit.MultiBeam(beams_file, group_name='j033216m2743',
                                MW_EBV=-1, fcontam=0.1, sys_err=0.03)
    
        assert mb.N == 2
        assert 'G102' in mb.PA
        
        expn, expt = mb.compute_exptime()
        assert np.allclose(expt['G102'], 1102.93, rtol=1.e-2)
        
        mb.compute_model()
        
        spec = mb.oned_spectrum()
        
    def test_redshift_fit(self):
        """
        Can we run the full fit?
        """
        path = os.path.dirname(utils.__file__)
        data_path = os.path.join(path, 'tests', 'data')
    
        res = fitting.run_all_parallel(152, zr=[1.7, 1.8], verbose=True,
                                 root=os.path.join(data_path, 'j033216m2743'),
                                 args_file=os.path.join(data_path, 'fit_args.npy'), 
                                 get_output_data=True, 
                                 protect=False)
        
        assert np.allclose(res[2].meta['z_map'][0], 1.7429, rtol=1.e-2)
            
        # Clean up
        files = glob.glob('j033216m2743_00152*')
        for file in files:
            os.remove(file)



"""
Tools for generating *very* basic fake images for HST/JWST/Roman simulations
"""
import os
import numpy as np

import astropy.io.fits as pyfits
import astropy.wcs as pywcs

from . import GRIZLI_PATH

def rotate_CD_matrix(cd, pa_aper):
    """
    Rotate CD matrix

    Parameters
    ----------
    cd : (2,2) array
        CD matrix

    pa_aper : float
        Position angle, in degrees E from N, of y axis of the detector

    Returns
    -------
    cd_rot : (2,2) array
        Rotated CD matrix

    Notes
    -----
    `astropy.wcs.WCS.rotateCD` doesn't work for non-square pixels in that it
    doesn't preserve the pixel scale!  The bug seems to come from the fact
    that `rotateCD` assumes a transposed version of its own CD matrix.

    For example:

        >>> import astropy.wcs as pywcs
        >>>
        >>> ## Nominal rectangular WFC3/IR pixel
        >>> cd_wfc3 = np.array([[  2.35945978e-05,   2.62448998e-05],
        >>>                     [  2.93050803e-05,  -2.09858771e-05]])
        >>>
        >>> ## Square pixel
        >>> cd_square = np.array([[0.1/3600., 0], [0, 0.1/3600.]])
        >>>
        >>> for cd, label in zip([cd_wfc3, cd_square], ['WFC3/IR', 'Square']):
        >>>     wcs = pywcs.WCS()
        >>>     wcs.wcs.cd = cd
        >>>     wcs.rotateCD(45.)
        >>>     print '%s pixel: pre=%s, rot=%s' %(label,
        >>>                         np.sqrt((cd**2).sum(axis=0))*3600,
        >>>                         np.sqrt((wcs.wcs.cd**2).sum(axis=0))*3600)

        WFC3/IR pixel:   pre=[ 0.1354  0.121 ], rot=[ 0.1282  0.1286]
        Square  pixel: pre=[ 0.1  0.1], rot=[ 0.1  0.1]

    """
    rad = np.deg2rad(-pa_aper)
    mat = np.zeros((2, 2))
    mat[0, :] = np.array([np.cos(rad), -np.sin(rad)])
    mat[1, :] = np.array([np.sin(rad), np.cos(rad)])
    cd_rot = np.dot(mat, cd)
    return cd_rot


def niriss_header(
    ra=53.1592277508136,
    dec=-27.782056346146,
    pa_aper=128.589,
    filter="F150W",
    grism="GR150R",
):
    """
    Make JWST/NIRISS image header

    Parameters
    ----------
    ra, dec : float, float
        Coordinates of the center of the image

    pa_aper : float
        Position angle of the y-axis of the detector

    filter : str
        Blocking filter to use.

    grism : str
        Grism to use

    Returns
    -------
    h : `astropy.io.fits.Header`
        FITS header with appropriate keywords

    wcs : `astropy.wcs.WCS`
        WCS specification (computed from keywords in ``h``).

    Notes
    -----
    NIRISS: 0.065"/pix, requires filter & grism specification
    """
    naxis = 2048, 2048
    crpix = 1024, 1024

    cd = np.array([[-0.0658, 0], [0, 0.0654]]) / 3600.0
    cd_rot = rotate_CD_matrix(cd, pa_aper)

    h = pyfits.Header()

    h["CRVAL1"] = ra
    h["CRVAL2"] = dec

    h["WCSAXES"] = 2
    h["CTYPE1"] = "RA---TAN"
    h["CTYPE2"] = "DEC--TAN"

    for i in range(2):
        h["NAXIS%d" % (i + 1)] = naxis[i]
        h["CRPIX%d" % (i + 1)] = crpix[i]
        h["CDELT%d" % (i + 1)] = 1.0
        for j in range(2):
            h["CD%d_%d" % (i + 1, j + 1)] = cd_rot[i, j]

    # Backgrounds
    # http://www.stsci.edu/jwst/instruments/niriss/software-tools/wfss-simulations/niriss-wfss-cookbook.pdf
    bg = {
        "F090W": 0.50,
        "F115W": 0.47,
        "F140M": 0.23,
        "F150W": 0.48,
        "F158M": 0.25,
        "F200W": 0.44,
    }

    h["INSTRUME"] = "NIRISS"
    h["TELESCOP"] = "JWST"
    h["DETECTOR"] = "NIS"

    if grism == "GR150R":
        h["FILTER"] = "GR150R", "Spectral trace along Y"
    else:
        h["FILTER"] = "GR150C", "Spectral trace along X"

    h["PUPIL"] = filter

    h["BACKGR"] = bg[filter], "Total, e/s"
    h["READN"] = 6, "Rough, per pixel per 1 ks exposure"  # e/pix/per
    h["PHOTFLAM"] = 1.0
    h["PHOTPLAM"] = 1.0

    wcs = pywcs.WCS(h)
    h["EXTVER"] = 1

    return h, wcs


def nircam_header(
    ra=53.1592277508136,
    dec=-27.782056346146,
    pa_aper=128.589,
    filter="F444W",
    grism="GRISMR",
    module="A",
):
    """
    Make JWST/NIRCAM image header

    Parameters
    ----------
    ra, dec : float, float
        Coordinates of the center of the image

    pa_aper : float
        Position angle of the y-axis of the detector

    filter : str
        Blocking filter to use.

    grism : str
        Grism to use ('GRISMR', 'GRISMC')

    module : str
        Instrument module ('A','B')

    Returns
    -------
    h : `astropy.io.fits.Header`
        FITS header with appropriate keywords

    wcs : `astropy.wcs.WCS`
        WCS specification (computed from keywords in ``h``).

    Notes
    -----
    NIRCAM, 0.0648"/pix, requires filter specification
    """
    naxis = 2048, 2048
    crpix = 1024, 1024

    cd = np.array([[-0.0648, 0], [0, 0.0648]]) / 3600.0
    cd_rot = rotate_CD_matrix(cd, pa_aper)

    h = pyfits.Header()

    h["CRVAL1"] = ra
    h["CRVAL2"] = dec

    h["WCSAXES"] = 2
    h["CTYPE1"] = "RA---TAN"
    h["CTYPE2"] = "DEC--TAN"

    for i in range(2):
        h["NAXIS%d" % (i + 1)] = naxis[i]
        h["CRPIX%d" % (i + 1)] = crpix[i]
        h["CDELT%d" % (i + 1)] = 1.0
        for j in range(2):
            h["CD%d_%d" % (i + 1, j + 1)] = cd_rot[i, j]

    # Backgrounds
    # http://www.stsci.edu/jwst/instruments/niriss/software-tools/wfss-simulations/niriss-wfss-cookbook.pdf
    bg = {
        "F277W": 0.30,
        "F356W": 0.90,
        "F444W": 3.00,
        "F322W2": 1.25,
        "F430M": 0.65,
        "F460M": 0.86,
        "F410M": 0.5,
    }  # F410M is a hack, no number

    h["BACKGR"] = bg[filter], "Total, e/s"
    h["INSTRUME"] = "NIRCAM"
    h["TELESCOP"] = "JWST"
    h["DETECTOR"] = f"NRC{module}LONG"
    h["MODULE"] = module
    h["CHANNEl"] = "LONG"

    if grism == "GRISMR":
        h["PUPIL"] = "GRISMR", "Spectral trace along X"
    else:
        h["PUPIL"] = "GRISMC", "Spectral trace along Y"

    h["FILTER"] = filter

    h["READN"] = 9, "Rough, per pixel per 1 ks exposure"  # e/pix/per
    h["PHOTFLAM"] = 1.0
    h["PHOTPLAM"] = 1.0

    wcs = pywcs.WCS(h)
    h["EXTVER"] = 1

    return h, wcs


def wfc3ir_header(
    ra=53.1592277508136,
    dec=-27.782056346146,
    pa_aper=128.589,
    flt="ibhj34h6q_flt.fits",
    filter="G141",
):
    """
    Make HST/WFC3-IR image header

    Parameters
    ----------
    ra, dec : float, float
        Coordinates of the center of the image

    pa_aper : float
        Position angle of the y-axis of the detector

    flt : str
        Filename of a WFC3/IR FLT file that will be used to provide the
        SIP geometric distortion keywords.

    filter : str
        Grism/filter to use.

    Returns
    -------
    h : `astropy.io.fits.Header`
        FITS header with appropriate keywords

    wcs : `astropy.wcs.WCS`
        WCS specification (computed from keywords in ``h``).

    Notes
    -----
    WFC3 IR, requires reference FLT file for the SIP header
    """
    import numpy as np

    import astropy.io.fits as pyfits
    import astropy.wcs as pywcs

    im = pyfits.open(flt)
    wcs = pywcs.WCS(im[1].header, relax=True)

    thet0 = np.arctan2(im[1].header["CD2_2"], im[1].header["CD2_1"]) / np.pi * 180

    wcs.wcs.crval = np.array([ra, dec])

    # Rotate the CD matrix
    theta = im[1].header["PA_APER"] - pa_aper
    cd_rot = rotate_CD_matrix(wcs.wcs.cd, theta)
    wcs.wcs.cd = cd_rot

    h = wcs.to_header(relax=True)

    for i in [1, 2]:
        for j in [1, 2]:
            h["CD%d_%d" % (i, j)] = h["PC%d_%d" % (i, j)]
            h.remove("PC%d_%d" % (i, j))

    h["BACKGR"] = 1.0
    h["FILTER"] = filter
    h["INSTRUME"] = "WFC3"
    h["READN"] = im[0].header["READNSEA"]
    h["NAXIS1"] = h["NAXIS2"] = 1014
    h["DETECTOR"] = "IR"
    h["PHOTFLAM"] = 1.0
    h["PHOTPLAM"] = 1.0

    return h, wcs


def wfirst_header(**kwargs):
    """
    Alias to `~grizli.fake_image.roman_header`
    """
    res = roman_header(**kwargs)
    return res


def roman_header(
    ra=53.1592277508136,
    dec=-27.782056346146,
    pa_aper=128.589,
    naxis=(4096, 4096),
    **kwargs,
):
    """
    Make WFIRST/Roman WFI header

    Parameters
    ----------
    ra, dec : float, float
        Coordinates of the center of the image

    pa_aper : float
        Position angle of the y-axis of the detector

    naxis : (int,int)
        Image dimensions

    Returns
    -------
    h : `astropy.io.fits.Header`
        FITS header with appropriate keywords

    wcs : `astropy.wcs.WCS`
        WCS specification (computed from keywords in `h`).

    Notes
    -----
    WFIRST/Roman G150 Grism

    Current config file has no field dependence, so field size can be
    anything you want in ``naxis``.
    """
    # naxis = 2048, 2048
    crpix = naxis[0] / 2.0, naxis[0] / 2.0

    cd = np.array([[-0.11, 0], [0, 0.11]]) / 3600.0
    cd_rot = rotate_CD_matrix(cd, pa_aper)

    h = pyfits.Header()

    h["CRVAL1"] = ra
    h["CRVAL2"] = dec

    h["WCSAXES"] = 2
    h["CTYPE1"] = "RA---TAN"
    h["CTYPE2"] = "DEC--TAN"

    for i in range(2):
        h["NAXIS%d" % (i + 1)] = naxis[i]
        h["CRPIX%d" % (i + 1)] = crpix[i]
        h["CDELT%d" % (i + 1)] = 1.0
        for j in range(2):
            h["CD%d_%d" % (i + 1, j + 1)] = cd_rot[i, j]

    # h['BACKGR'] = 0.17+0.49, 'Total, e/s SDT Report A-1'
    h["BACKGR"] = 1.12, "Pandeia minzodi/benchmark 20210528"
    h["FILTER"] = "G150", "WFIRST/Roman grism"
    h["INSTRUME"] = "WFI"
    # h['READN'] = 17, 'SDT report Table 3-3'  # e/pix/per

    # https://roman.gsfc.nasa.gov/science/RRI/Roman_WFI_Reference_Information_20210125.pdf
    h["READN"] = 16.0, "WFI Reference 20210125"  # e/pix/per
    h["PHOTFLAM"] = 1.0
    h["PHOTPLAM"] = 1.0

    wcs = pywcs.WCS(h)
    h["EXTVER"] = 1

    return h, wcs


def roman_hls_image(
    exptime=661.932, nexp=13, background=1.12, output="roman.fits", **kwargs
):
    """
    Make a simple FITS file for a Roman High Latitude Survey Image

    Parameters
    ----------
    exptime, nexp, background : float, int, float
        Defaults specified to roughly match the variance in the `pandeia`
        2D simulation result (ignoring Poisson from the source)

    output : str
        Output filename passed to `~grizli.fake_image.make_fake_image`.
        Default is 'roman.fits'.

    kwargs : dict
        Positional keywords passed to `~grizli.fake_image.roman_header`

    Returns
    -------
    hdu : `astropy.io.fits.HDUList`
        HDU with SCI, ERR, DQ extensions

    wcs : `astropy.wcs.WCS`
        WCS
    """
    header, wcs = roman_header(**kwargs)
    hdu = make_fake_image(
        header, output=output, background=background, exptime=exptime, nexp=nexp
    )

    return hdu, wcs


def make_fake_image(
    header,
    output="direct.fits",
    background=None,
    exptime=1.0e4,
    nexp=10,
    obsdate=None,
    seed=None,
):
    """
    Use the header from NIRISS, WFC3/IR or WFIRST/Roman and make an ``FLT``-like image that `grizli` can read as a reference.

    Parameters
    ----------
    header : `astropy.io.fits.Header`
        Header created by one of the generating functions, such as
        `~grizli.fake_image.niriss_header`.

    output : str
        Filename of the output FITS file. Will have extensions 'SCI', 'ERR',
        and 'DQ'. The 'ERR' extension is populated with a read-noise +
        background error model using

            >>> var = nexp*header['READN'] + background*exptime

        The 'SCI' extension is filled with gaussian deviates with standard
        deviation `sqrt(var)`.

        The 'DQ' extension is filled with (int) zeros.

    background : None or float
        Background value to use for sky noise.  If None, then read from
        `header['BACKGR']`.

    exptime : float
        Exposure time to use for background sky noise.

    nexp : int
        Number of exposures to use for read noise.

    obsdate : `~astropy.time.Time`
        Date of observation.  If None, then use `astropy.time.Time.now`

    seed : int
        If specified, use as `numpy.random.seed`

    Returns
    -------
    hdu : `astropy.io.fits.HDUList`
        Image HDU (also saved to ``output`` FITS file)

    """
    import astropy.time
    import astropy.units as u

    hdu = pyfits.HDUList()

    header["EXPTIME"] = exptime
    header["NEXP"] = nexp
    header["BUNIT"] = "ELECTRONS/S"

    hdu.append(pyfits.PrimaryHDU(header=header))

    naxis = (header["NAXIS1"], header["NAXIS2"])

    if background is None:
        background = header["BACKGR"]

    header["BACKGR"] = background

    if obsdate is None:
        obsdate = astropy.time.Time.now()

    header["DATE-OBS"] = obsdate.iso.split()[0]
    header["TIME-OBS"] = obsdate.iso.split()[1]
    header["EXPSTART"] = obsdate.mjd
    header["EXPEND"] = (obsdate + exptime * u.second).mjd

    # Simple error model of read noise and sky background
    var = nexp * header["READN"] + background * exptime

    # electrons / s
    rms = np.sqrt(var) / exptime
    header["CALCRMS"] = rms, "Variance used for random noise"

    for name, dtype in zip(["SCI", "ERR", "DQ"], [np.float32, np.float32, np.int32]):
        hdu.append(
            pyfits.ImageHDU(
                header=header, data=np.zeros(np.array(naxis).T, dtype=dtype), name=name
            )
        )

    hdu["ERR"].data += rms
    if seed is not None:
        np.random.seed(seed)
        hdu["ERR"].header["SEED"] = seed, "Random seed"

    hdu["SCI"].data = np.random.normal(size=np.array(naxis).T) * rms

    if output is not None:
        hdu.writeto(output, overwrite=True, output_verify="fix")

    return hdu


def make_roman_config(save_to_conf=True):
    """
    Use `pandeia` to calculate a Roman/G150 configuration file and sensitivity curve for `grizli`

    https://github.com/spacetelescope/roman_tools/blob/develop/notebooks/Pandeia-Roman.ipynb

    Parameters
    ----------
    save_to_conf : bool
        Write sensitivity and configuration files to ``[GRIZLI_PATH]/CONF``

    Returns
    -------
    sens : `~astropy.table.Table`
        Sensitivity table

    conf : str
        Grism configuration

    """
    from astropy.table import Table
    import astropy.time
    import pandeia.engine

    from pandeia.engine.perform_calculation import perform_calculation
    from pandeia.engine.calc_utils import (
        get_telescope_config,
        get_instrument_config,
        build_default_calc,
        build_default_source,
    )

    from pandeia.engine.io_utils import read_json, write_json

    calc = build_default_calc("roman", "wfi", "spectroscopy")

    # HLS simulation
    calc["configuration"]["instrument"]["filter"] = None
    calc["configuration"]["instrument"]["aperture"] = "any"
    calc["configuration"]["instrument"]["disperser"] = "g150"
    calc["configuration"]["detector"]["ngroup"] = 13  # groups per integration
    calc["configuration"]["detector"]["nint"] = 1  # integrations per exposure
    calc["configuration"]["detector"]["nexp"] = 1  # exposures
    calc["configuration"]["detector"]["readmode"] = "medium8"
    calc["configuration"]["detector"]["subarray"] = "1024x1024"

    calc["scene"][0]["spectrum"]["normalization"]["norm_fluxunit"] = "flam"

    input_flux = 1.0e-19
    calc["scene"][0]["spectrum"]["normalization"]["norm_flux"] = input_flux
    calc["scene"][0]["spectrum"]["sed"]["unit"] = "flam"

    # x,y location to extract, in arcsec
    calc["strategy"]["target_xy"] = [0.0, 0.0]
    # radius of extraction aperture, in arcsec
    calc["strategy"]["aperture_size"] = 0.6
    # inner and outer radii of background subtraction annulus, in arcsec
    calc["strategy"]["sky_annulus"] = [0.8, 1.2]

    results = perform_calculation(calc)

    sens = Table()
    wave = np.arange(9000, 1.95e4, 2.0)
    sens_value = results["1d"]["extracted_flux"][1] / input_flux
    sens_value /= np.gradient(results["1d"]["extracted_flux"][0] * 1.0e4)

    sens["WAVELENGTH"] = wave
    sens["SENSITIVITY"] = np.interp(
        wave, results["1d"]["extracted_flux"][0] * 1.0e4, sens_value, left=0, right=0
    )

    sens["ERROR"] = 0.01 * sens_value.max()

    sens.meta["pandeia"] = pandeia.engine.__version__
    sens.meta["created"] = astropy.time.Time.now().iso

    sens_file = f"Roman.G150.v{pandeia.engine.__version__}.sens.fits"

    if save_to_conf:
        if isinstance(save_to_conf, str):
            path = save_to_conf
        else:
            path = os.path.join(GRIZLI_PATH, "CONF")

        print("Sensitivity file: ", os.path.join(path, sens_file))
        sens.write(os.path.join(path, sens_file), overwrite=True)

    npix = len(results["1d"]["extracted_flux"][0])
    pad = 20

    i0 = npix // 2
    w0 = results["1d"]["extracted_flux"][0][i0] * 1.0e4
    dlam = np.diff(results["1d"]["extracted_flux"][0])[i0] * 1.0e4

    config = f"""INSTRUMENT WFI
GFILTER G150 

# First order (BEAM A) 
# BEAMA and DLDP assume spectrum is centered on the imaging position
BEAMA {-npix//2-pad} {npix//2+pad+1}
MMAG_EXTRACT_A 30
MMAG_MARK_A 30

#
# Trace description
# (flat along rows)
DYDX_ORDER_A 0
DYDX_A_0 0

#
# X and Y Offsets
#
XOFF_A 0.0
YOFF_A 0.0

#
# Dispersion solution
#
DISP_ORDER_A 1
DLDP_A_0 {w0}
DLDP_A_1 {dlam}

SENSITIVITY_A {sens_file}
"""

    if save_to_conf:
        print("Config file: ", os.path.join(path, "Roman.G150.conf"))
        with open(os.path.join(path, "Roman.G150.conf"), "w") as fp:
            fp.write(config)

    return sens, config